from langchain_graph_retriever.retrievers.strategy.mmr import Mmr


def test_mmr_parameters() -> None:
    # Makes sure that copying the MMR startegy creates new embeddings.
    mmr1 = Mmr(query_embedding=[0.25, 0.5, 0.75])
    assert id(mmr1._nd_query_embedding) == id(mmr1._nd_query_embedding)

    mmr2 = mmr1.model_copy(deep=True)
    assert id(mmr1._nd_query_embedding) != id(mmr2._nd_query_embedding)



import asyncio

import pytest
from pytest import Parser
from tests.integration_tests.stores import ALL_STORES, TESTCONTAINER_STORES


@pytest.fixture(scope="session")
def event_loop():
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
    yield loop
    loop.close()


def pytest_addoption(parser: Parser):
    parser.addoption(
        "--stores",
        action="append",
        metavar="STORE",
        choices=ALL_STORES + ["all"],
        help="run tests for the given store (default: 'mem' and 'mem_norm')",
    )
    parser.addoption(
        "--testcontainer",
        action="append",
        metavar="STORE",
        choices=TESTCONTAINER_STORES + ["none"],
        help="which stores to run testcontainers for (default: 'all')",
    )


# TODO: Allow marking tests as only supporting a subset of stores?
#
# def pytest_configure(config):
#     # register an additional marker
#     config.addinivalue_line(
#         "markers", "svc(name): tests that require the named service"
#     )
#
# def pytest_runtest_setup(item):
#     """Skip the test unless all of the marked services are present."""
#
#     required_svcs = {mark.args[0] for mark in item.iter_markers(name="svc")}
#     provided_svcs = set(item.config.getoption("-S") or [])
#
#     missing_svcs = required_svcs - provided_svcs
#     if missing_svcs:
#         pytest.skip(f"test requires services {missing_svcs!r}")



import asyncio
from typing import (
    Any,
    Iterable,
    List,
    Sequence,
    Tuple,
    Union,
)

from langchain_core.documents import Document
from langchain_core.retrievers import BaseRetriever
from pydantic import Field, computed_field

from .edge import Edge
from .edge_helper import EdgeHelper
from .node import Node
from .store_adapters.base import METADATA_EMBEDDING_KEY, StoreAdapter
from .strategy.base import TraversalStrategy

INFINITY = float("inf")


class _TraversalState:
    """Manages the book-keeping necessary to keep track of traversal state."""

    def __init__(
        self,
        *,
        edge_helper: EdgeHelper,
        base_strategy: TraversalStrategy | None,
        strategy: TraversalStrategy | dict[str, Any] | None,
    ) -> None:
        self.edge_helper = edge_helper

        # Deep copy in case the strategy has mutable state
        if isinstance(strategy, TraversalStrategy):
            self.strategy = strategy.model_copy(deep=True)
        elif isinstance(strategy, dict):
            assert (
                base_strategy is not None
            ), "Must set strategy in init to support field-overrides."
            self.strategy = base_strategy.model_copy(update=strategy, deep=True)
        elif strategy is None:
            assert base_strategy is not None, "Must set strategy in init or invocation."
            self.strategy = base_strategy.model_copy(deep=True)
        else:
            raise ValueError(f"Unsupported strategy {strategy}")

        self.visited_edges: set[Edge] = set()
        self.edge_depths: dict[Edge, int] = {}
        self.doc_cache: dict[str, Document] = {}
        self.node_cache: dict[str, Node] = {}

        self.selected_nodes: dict[str, Node] = {}

    def _doc_to_new_node(
        self, doc: Document, *, depth: int | None = None
    ) -> Node | None:
        if doc.id is None:
            raise ValueError("All documents should have ids")
        if doc.id in self.node_cache:
            return None

        doc = self.doc_cache.setdefault(doc.id, doc)
        assert doc.id is not None
        incoming_edges, outgoing_edges = self.edge_helper.get_incoming_outgoing(
            doc.metadata
        )
        if depth is None:
            depth = min(
                [
                    d
                    for e in incoming_edges
                    if (d := self.edge_depths.get(e, None)) is not None
                ],
                default=0,
            )
        node = Node(
            id=doc.id,
            depth=depth,
            embedding=doc.metadata[METADATA_EMBEDDING_KEY],
            metadata=doc.metadata,
            incoming_edges=incoming_edges,
            outgoing_edges=outgoing_edges,
        )
        self.node_cache[doc.id] = node

        return node

    def add_docs(
        self, docs: Iterable[Document], *, depth: int | None = None
    ) -> dict[str, Node]:
        # Record the depth of new nodes.
        nodes = {
            node.id: node
            for doc in docs
            if (node := self._doc_to_new_node(doc, depth=depth)) is not None
            if (
                self.strategy.max_depth is None or node.depth <= self.strategy.max_depth
            )
        }
        self.strategy.add_nodes(nodes)
        return nodes

    def visit_nodes(self, nodes: Iterable[Node]) -> set[Edge]:
        """Record the nodes as visited, returning the new outgoing edges.

        After this call, the outgoing edges will be added to the visited
        set, and not revisited during the traversal.
        """
        new_outgoing_edges: dict[Edge, int] = {}
        for node in nodes:
            node_new_outgoing_edges = node.outgoing_edges - self.visited_edges
            for edge in node_new_outgoing_edges:
                depth = new_outgoing_edges.setdefault(edge, node.depth + 1)
                if node.depth + 1 < depth:
                    new_outgoing_edges[edge] = node.depth + 1

        self.edge_depths.update(new_outgoing_edges)

        new_outgoing_edge_set = set(new_outgoing_edges.keys())
        self.visited_edges.update(new_outgoing_edge_set)
        return new_outgoing_edge_set

    def select_next_edges(self) -> set[Edge] | None:
        """Select the next round of nodes.

        Returns the set of new edges that need to be explored.
        """
        remaining = self.strategy.k - len(self.selected_nodes)

        if remaining <= 0:
            return None

        next_nodes = self.strategy.select_nodes(limit=remaining)
        if not next_nodes:
            return None

        next_nodes = [n for n in next_nodes if n.id not in self.selected_nodes]
        if len(next_nodes) == 0:
            return None

        self.selected_nodes.update({n.id: n for n in next_nodes})
        new_outgoing_edges = self.visit_nodes(next_nodes)
        return new_outgoing_edges

    def finish(self) -> list[Document]:
        final_nodes = self.strategy.finalize_nodes(self.selected_nodes.values())
        docs = []
        for node in final_nodes:
            doc = self.doc_cache.get(node.id, None)
            if doc is None:
                raise RuntimeError(
                    f"unexpected, cache should contain doc id: {node.id}"
                )
            # Compute new metadata from extra metadata and metadata.
            # This allows us to avoid modifying the orginal metadata.
            metadata = {
                "depth": node.depth,
                **node.extra_metadata,
                **doc.metadata,
            }
            # Remove the metadata embedding key. TODO: Find a better way to do this.
            metadata.pop(METADATA_EMBEDDING_KEY, None)
            docs.append(
                Document(
                    id=node.id,
                    page_content=doc.page_content,
                    metadata=metadata,
                )
            )
        return docs


# this class uses pydantic, so store and edges
# must be provided at init time.
class GraphTraversalRetriever(BaseRetriever):
    store: StoreAdapter
    edges: List[Union[str, Tuple[str, str]]]
    strategy: TraversalStrategy | None = None

    denormalized_path_delimiter: str = Field(default=".")
    denormalized_static_value: Any = Field(default="$")

    extra_args: dict[str, Any] = {}

    @computed_field  # type: ignore
    @property
    def edge_helper(self) -> EdgeHelper:
        return EdgeHelper(
            edges=self.edges,
            denormalized_path_delimiter=self.denormalized_path_delimiter,
            denormalized_static_value=self.denormalized_static_value,
            use_normalized_metadata=self.store.supports_normalized_metadata,
        )

    def _get_relevant_documents(
        self,
        query: str,
        *,
        strategy: TraversalStrategy | dict[str, Any] | None = None,
        initial_roots: Sequence[str] = (),
        filter: dict[str, Any] | None = None,
        store_kwargs: dict[str, Any] = {},
        **kwargs: Any,
    ) -> list[Document]:
        """Retrieve document nodes from this graph vector store using MMR-traversal.
        This strategy first retrieves the top `start_k` results by similarity to
        the question. It then selects the top `k` results based on
        maximum-marginal relevance using the given `lambda_mult`.
        At each step, it considers the (remaining) documents from `start_k` as
        well as any documents connected by edges to a selected document
        retrieved based on similarity (a "root").
        Args:
            query: The query string to search for.
            strategy: Specify or override the strategy to use for this retrieval.
            initial_roots: Optional list of document IDs to use for initializing search.
                The top `adjacent_k` nodes adjacent to each initial root will be
                included in the set of initial candidates. To fetch only in the
                neighborhood of these nodes, set `start_k = 0`.
            filter: Optional metadata to filter the results.
            store_kwargs: Optional kwargs passed to queries to the store.
            **kwargs: Additional keyword arguments passed to traversal state.
        """
        state = _TraversalState(
            base_strategy=self.strategy,
            strategy=strategy,
            edge_helper=self.edge_helper,
        )

        # Retrieve initial candidates.
        initial_docs = self._fetch_initial_candidates(
            query, state=state, filter=filter, **store_kwargs
        )
        state.add_docs(initial_docs, depth=0)

        if initial_roots:
            neighborhood_adjacent_docs = self._fetch_neighborhood_candidates(
                initial_roots,
                state=state,
                filter=filter,
                **store_kwargs,
            )
            state.add_docs(neighborhood_adjacent_docs, depth=0)

        while True:
            # Select the next batch of nodes, and (new) outgoing edges.
            next_outgoing_edges = state.select_next_edges()
            if next_outgoing_edges is None:
                break
            elif next_outgoing_edges:
                # Find the (new) document with incoming edges from those edges.
                adjacent_docs = self._get_adjacent(
                    outgoing_edges=next_outgoing_edges,
                    state=state,
                    filter=filter,
                    **store_kwargs,
                )

                state.add_docs(adjacent_docs)

        return state.finish()

    async def _aget_relevant_documents(
        self,
        query: str,
        *,
        strategy: TraversalStrategy | dict[str, Any] | None = None,
        initial_roots: Sequence[str] = (),
        filter: dict[str, Any] | None = None,
        store_kwargs: dict[str, Any] = {},
        **kwargs: Any,
    ) -> list[Document]:
        """Asynchronously retrieve documents from this graph store using MMR-traversal.

        This strategy first retrieves the top `start_k` results by similarity to
        the question. It then selects the top `k` results based on
        maximum-marginal relevance using the given `lambda_mult`.
        At each step, it considers the (remaining) documents from `start_k` as
        well as any documents connected by edges to a selected document
        retrieved based on similarity (a "root").

        Args:
            query: The query string to search for.
            strategy: Specify or override the strategy to use for this retrieval.
            initial_roots: Optional list of document IDs to use for initializing search.
                The top `adjacent_k` nodes adjacent to each initial root will be
                included in the set of initial candidates. To fetch only in the
                neighborhood of these nodes, set `start_k = 0`.
            filter: Optional metadata to filter the results.
            store_kwargs: Optional kwargs passed to queries to the store.
            **kwargs: Additional keyword arguments passed to traversal state.
        """
        state = _TraversalState(
            edge_helper=self.edge_helper,
            base_strategy=self.strategy,
            strategy=strategy,
        )

        # Retrieve initial candidates and initialize state.
        initial_docs = await self._afetch_initial_candidates(
            query, state=state, filter=filter, **store_kwargs
        )
        state.add_docs(initial_docs, depth=0)

        if initial_roots:
            neighborhood_adjacent_docs = await self._afetch_neighborhood_candidates(
                initial_roots,
                state=state,
                filter=filter,
                **store_kwargs,
            )
            state.add_docs(neighborhood_adjacent_docs, depth=0)

        while True:
            # Select the next batch of nodes, and (new) outgoing edges.
            next_outgoing_edges = state.select_next_edges()
            if next_outgoing_edges is None:
                break
            elif next_outgoing_edges:
                # Find the (new) document with incoming edges from those edges.
                adjacent_docs = await self._aget_adjacent(
                    outgoing_edges=next_outgoing_edges,
                    state=state,
                    filter=filter,
                    **store_kwargs,
                )

                state.add_docs(adjacent_docs)

        return state.finish()

    def _fetch_initial_candidates(
        self,
        query: str,
        *,
        state: _TraversalState,
        filter: dict[str, Any] | None,
        **kwargs: Any,
    ) -> Iterable[Document]:
        """Gets the embedded query and the set of initial candidates.

        Args:
            query: String to compute embedding and fetch initial matches for.
            state: The travel state we're retrieving candidates fore.
            filter: Optional metadata filter to apply.
            **kwargs: Additional keyword arguments.
        """
        query_embedding, docs = self.store.similarity_search_with_embedding(
            query=query,
            k=state.strategy.start_k,
            filter=filter,
            **kwargs,
        )
        state.strategy.query_embedding = query_embedding
        return docs

    async def _afetch_initial_candidates(
        self,
        query: str,
        *,
        state: _TraversalState,
        filter: dict[str, Any] | None,
        **kwargs: Any,
    ) -> Iterable[Document]:
        query_embedding, docs = await self.store.asimilarity_search_with_embedding(
            query=query,
            k=state.strategy.start_k,
            filter=filter,
            **kwargs,
        )
        state.strategy.query_embedding = query_embedding
        return docs

    def _fetch_neighborhood_candidates(
        self,
        neighborhood: Sequence[str],
        *,
        state: _TraversalState,
        filter: dict[str, Any] | None,
        **kwargs: Any,
    ) -> Iterable[Document]:
        neighborhood_docs = self.store.get(neighborhood)
        neighborhood_nodes = state.add_docs(neighborhood_docs)

        # Record the neighborhood nodes (specifically the outgoing edges from the
        # neighborhood) as visited.
        outgoing_edges = state.visit_nodes(neighborhood_nodes.values())

        # Fetch the candidates.
        return self._get_adjacent(
            outgoing_edges=outgoing_edges,
            query_embedding=state.strategy.query_embedding,
            k_per_edge=state.strategy.adjacent_k,
            filter=filter,
            **kwargs,
        )

    async def _afetch_neighborhood_candidates(
        self,
        neighborhood: Sequence[str],
        *,
        query_embedding: list[float],
        adjacent_k: int,
        state: _TraversalState,
        filter: dict[str, Any] | None,
        **kwargs: Any,
    ):
        neighborhood_docs = await self.store.aget(neighborhood)
        neighborhood_nodes = state.add_docs(neighborhood_docs)

        # Record the neighborhood nodes (specifically the outgoing edges from the
        # neighborhood) as visited.
        outgoing_edges = state.visit_nodes(neighborhood_nodes.values())

        # Fetch the candidates.
        return await self._aget_adjacent(
            outgoing_edges=outgoing_edges,
            query_embedding=query_embedding,
            k_per_edge=adjacent_k,
            filter=filter,
            **kwargs,
        )

    def _get_adjacent(
        self,
        outgoing_edges: set[Edge],
        state: _TraversalState,
        filter: dict[str, Any] | None,
        **kwargs: Any,
    ) -> Iterable[Document]:
        """Return the target docs with incoming edges from any of the given edges.
        Args:
            edges: The edges to look for.
            state: Traversal state we're retrieving adjacent nodes for.
            filter: Metadata to filter the results.
        Returns:
            Dictionary of adjacent nodes, keyed by node ID.
        """
        results: list[Document] = []
        for outgoing_edge in outgoing_edges:
            docs = self.store.similarity_search_with_embedding_by_vector(
                embedding=state.strategy.query_embedding,
                k=state.strategy.adjacent_k,
                filter=self.edge_helper.get_metadata_filter(
                    base_filter=filter, edge=outgoing_edge
                ),
                **kwargs,
            )
            results.extend(docs)
            if not self.store.supports_normalized_metadata:
                # If we denormalized the metadata, we actually do two queries.
                # One, for normalized values (above) and one for denormalized.
                # This ensures that cases where the key had a single value are
                # caught as well. This could *maybe* be handled differently if
                # we know keys that were always denormalized.
                docs = self.store.similarity_search_with_embedding_by_vector(
                    embedding=state.strategy.query_embedding,
                    k=state.strategy.adjacent_k,
                    filter=self.edge_helper.get_metadata_filter(
                        base_filter=filter, edge=outgoing_edge, denormalize_edge=True
                    ),
                    **kwargs,
                )
                results.extend(docs)
        return results

    async def _aget_adjacent(
        self,
        outgoing_edges: set[Edge],
        state: _TraversalState,
        filter: dict[str, Any] | None,
        **kwargs: Any,
    ) -> Iterable[Document]:
        """Returns document nodes with incoming edges from any of the given edges.
        Args:
            edges: The edges to look for.
            query_embedding: The query embedding. Used to rank target nodes.
            k_per_edge: The number of target nodes to fetch for each edge.
            filter: Optional metadata to filter the results.
        Returns:
            Dictionary of adjacent nodes, keyed by node ID.
        """

        tasks = [
            self.store.asimilarity_search_with_embedding_by_vector(
                embedding=state.strategy.query_embedding,
                k=state.strategy.adjacent_k,
                filter=self.edge_helper.get_metadata_filter(
                    base_filter=filter, edge=outgoing_edge
                ),
                **kwargs,
            )
            for outgoing_edge in outgoing_edges
        ]
        if not self.store.supports_normalized_metadata:
            # If we denormalized the metadata, we actually do two queries.
            # One, for normalized values (above) and one for denormalized.
            # This ensures that cases where the key had a single value are
            # caught as well. This could *maybe* be handled differently if
            # we know keys that were always denormalized.
            tasks.extend(
                [
                    self.store.asimilarity_search_with_embedding_by_vector(
                        embedding=state.strategy.query_embedding,
                        k=state.strategy.adjacent_k,
                        filter=self.edge_helper.get_metadata_filter(
                            base_filter=filter,
                            edge=outgoing_edge,
                            denormalize_edge=True,
                        ),
                        **kwargs,
                    )
                    for outgoing_edge in outgoing_edges
                ]
            )

        results: list[Document] = []
        for completed_task in asyncio.as_completed(tasks):
            docs = await completed_task
            results.extend(docs)
        return results



import abc
from typing import (
    Any,
    Dict,
    Generic,
    List,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
)

from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings
from langchain_core.runnables import run_in_executor
from langchain_core.vectorstores import VectorStore

StoreT = TypeVar("StoreT", bound=VectorStore)

METADATA_EMBEDDING_KEY = "__embedding"


class StoreAdapter(Generic[StoreT], abc.ABC):
    def __init__(self, vector_store: StoreT):
        self.vector_store = vector_store

    @property
    def _safe_embedding(self) -> Embeddings:
        if not self.vector_store.embeddings:
            msg = "Missing embedding"
            raise ValueError(msg)
        return self.vector_store.embeddings

    @property
    @abc.abstractmethod
    def supports_normalized_metadata(self) -> bool:
        """Returns ``True`` if the associated vector_store supports
        normalized metadata storage and filtering."""
        pass

    def similarity_search_with_embedding(
        self,
        query: str,
        k: int = 4,
        filter: Optional[Dict[str, str]] = None,
        **kwargs: Any,
    ) -> Tuple[List[float], List[Document]]:
        """Returns docs (with embeddings) most similar to the query.
        Also returns the embedded query vector.
        Args:
            query: Input text.
            k: Number of Documents to return. Defaults to 4.
            filter: Filter on the metadata to apply.
            **kwargs: Additional keyword arguments.
        Returns:
            A tuple of:
                * The embedded query vector
                * List of Documents most similar to the query vector.
                  Documents should have their embedding added to
                  their metadata under the METADATA_EMBEDDING_KEY key.
        """
        query_embedding = self._safe_embedding.embed_query(text=query)
        docs = self.similarity_search_with_embedding_by_vector(
            embedding=query_embedding,
            k=k,
            filter=filter,
            **kwargs,
        )
        return query_embedding, docs

    async def asimilarity_search_with_embedding(
        self,
        query: str,
        k: int = 4,
        filter: Optional[Dict[str, str]] = None,
        **kwargs: Any,
    ) -> Tuple[List[float], List[Document]]:
        """Returns docs (with embeddings) most similar to the query.
        Also returns the embedded query vector.
        Args:
            query: Input text.
            k: Number of Documents to return. Defaults to 4.
            filter: Filter on the metadata to apply.
            **kwargs: Additional keyword arguments.
        Returns:
            A tuple of:
                * The embedded query vector
                * List of Documents most similar to the query vector.
                  Documents should have their embedding added to
                  their metadata under the METADATA_EMBEDDING_KEY key.
        """
        return await run_in_executor(
            None, self.similarity_search_with_embedding, query, k, filter, **kwargs
        )

    @abc.abstractmethod
    def similarity_search_with_embedding_by_vector(
        self,
        embedding: List[float],
        k: int = 4,
        filter: Optional[Dict[str, str]] = None,
        **kwargs: Any,
    ) -> List[Document]:
        """Returns docs (with embeddings) most similar to the query vector.
        Args:
            embedding: Embedding to look up documents similar to.
            k: Number of Documents to return. Defaults to 4.
            filter: Filter on the metadata to apply.
            **kwargs: Additional keyword arguments.
        Returns:
            List of Documents most similar to the query vector.
                Documents should have their embedding added to
                their metadata under the METADATA_EMBEDDING_KEY key.
        """

    async def asimilarity_search_with_embedding_by_vector(
        self,
        embedding: List[float],
        k: int = 4,
        filter: Optional[Dict[str, str]] = None,
        **kwargs: Any,
    ) -> List[Document]:
        """Returns docs (with embeddings) most similar to the query vector.
        Args:
            embedding: Embedding to look up documents similar to.
            k: Number of Documents to return. Defaults to 4.
            filter: Filter on the metadata to apply.
            **kwargs: Additional keyword arguments.
        Returns:
            List of Documents most similar to the query vector.
                Documents should have their embedding added to
                their metadata under the METADATA_EMBEDDING_KEY key.
        """
        return await run_in_executor(
            None,
            self.similarity_search_with_embedding_by_vector,
            embedding,
            k,
            filter,
            **kwargs,
        )

    @abc.abstractmethod
    def get(
        self,
        ids: Sequence[str],
        /,
        **kwargs: Any,
    ) -> list[Document]:
        """Get documents by id.
        Fewer documents may be returned than requested if some IDs are not found or
        if there are duplicated IDs.
        Users should not assume that the order of the returned documents matches
        the order of the input IDs. Instead, users should rely on the ID field of the
        returned documents.
        This method should **NOT** raise exceptions if no documents are found for
        some IDs.

        Args:
            ids: List of IDs to get.
            kwargs: Additional keyword arguments. These are up to the implementation.
        Returns:
            List[Document]: List of documents that were found.
        """

    async def aget(
        self,
        ids: Sequence[str],
        /,
        **kwargs: Any,
    ) -> list[Document]:
        """Get documents by id.
        Fewer documents may be returned than requested if some IDs are not found or
        if there are duplicated IDs.
        Users should not assume that the order of the returned documents matches
        the order of the input IDs. Instead, users should rely on the ID field of the
        returned documents.
        This method should **NOT** raise exceptions if no documents are found for
        some IDs.

        Args:
            ids: List of IDs to get.
            kwargs: Additional keyword arguments. These are up to the implementation.
        Returns:
            List[Document]: List of documents that were found.
        """
        return await run_in_executor(
            None,
            self.get,
            ids,
            **kwargs,
        )



import heapq
from typing import Callable, Iterable

from ..node import Node
from .base import TraversalStrategy


class Scored(TraversalStrategy):
    """Node selector choosing the top `select_k` nodes according to `scorer`
    in each iteration.
    """

    scorer: Callable[[Node], float]
    """Scoring function to apply to each node.

    This will be invoked once when the node is first discovered, meaning
    the depth may be an upper-bound on the actual shortest path for the node.
    """

    select_k: int = 10
    """Number of top-scored nodes to select in each iteration. Default 10."""

    _nodes: list[tuple[float, Node]] = []

    def add_nodes(self, nodes: dict[str, Node]) -> None:
        for node in nodes.values():
            heapq.heappush(self._nodes, (self.scorer(node), node))

    def select_nodes(self, *, limit: int) -> Iterable[Node]:
        selected: list[Node] = []
        for _ in range(0, min(limit, self.select_k)):
            if len(self._nodes) == 0:
                break
            selected.append(heapq.heappop(self._nodes)[1])
        return selected



from dataclasses import dataclass, field
from typing import Any

from .edge import Edge


@dataclass
class Node:
    """Information about a node during the traversal."""

    id: str
    """The document ID of this node."""

    depth: int
    """The depth (number of edges) through which this node was discovered.
    This may be larger than the *actual* depth of the node in the complete
    graph. If only a subset of edges are retrieved (such as when using
    similarity search to select only the most relevant edges), then this will
    correspond to the depth in the retrieved set of edges, which may be more
    than the true depth, if edges necessary for the shorter path are not used.
    """

    embedding: list[float]

    incoming_edges: set[Edge]
    outgoing_edges: set[Edge]

    metadata: dict[str, Any] = field(default_factory=dict)
    """Metadata from the original document.

    This is a reference to the original document metadata, and should not be modified.
    Instead, modify `extra_metadata` which will be used as overrides when the final
    document is produced.
    """

    extra_metadata: dict[str, Any] = field(default_factory=dict)
    """Metadata to add to the original document for the results."""



import json
import math
import random
from abc import abstractmethod

from langchain_core.embeddings import Embeddings


class BaseEmbeddings(Embeddings):
    @abstractmethod
    def embed_query(self, text: str) -> list[float]:
        pass

    def embed_documents(self, texts: list[str]) -> list[list[float]]:
        return [self.embed_query(txt) for txt in texts]


class Angular2DEmbeddings(BaseEmbeddings):
    """
    From angles (as strings in units of pi) to unit embedding vectors on a circle.
    """

    def embed_query(self, text: str) -> list[float]:
        """
        Convert input text to a 'vector' (list of floats).
        If the text is a number, use it as the angle for the
        unit vector in units of pi.
        Any other input text becomes the singular result [0, 0] !
        """
        try:
            angle = float(text)
            return [math.cos(angle * math.pi), math.sin(angle * math.pi)]
        except ValueError:
            # Assume: just test string, no attention is paid to values.
            return [0.0, 0.0]


class EarthEmbeddings(BaseEmbeddings):
    def get_vector_near(self, value: float) -> list[float]:
        base_point = [value, (1 - value**2) ** 0.5]
        fluctuation = random.random() / 100.0
        return [base_point[0] + fluctuation, base_point[1] - fluctuation]

    def embed_query(self, text: str) -> list[float]:
        words = set(text.lower().split())
        if "earth" in words:
            vector = self.get_vector_near(0.9)
        elif {"planet", "world", "globe", "sphere"}.intersection(words):
            vector = self.get_vector_near(0.8)
        else:
            vector = self.get_vector_near(0.1)
        return vector


class ParserEmbeddings(BaseEmbeddings):
    """Parse input texts: if they are json for a List[float], fine.
    Otherwise, return all zeros and call it a day.
    """

    def __init__(self, dimension: int) -> None:
        self.dimension = dimension

    def embed_query(self, text: str) -> list[float]:
        try:
            vals = json.loads(text)
        except json.JSONDecodeError:
            return [0.0] * self.dimension
        else:
            assert len(vals) == self.dimension
            return vals


def string_to_number(word: str) -> int:
    return sum(ord(char) for char in word)


class WordEmbeddings(BaseEmbeddings):
    def __init__(self, words: list[str]):
        self._words = words
        self._offsets = [
            string_to_number(word=word) * ((-1) ** i) for i, word in enumerate(words)
        ]

    def embed_query(self, text: str) -> list[float]:
        return [
            1.0 + (100 / self._offsets[i]) if word in text else 0.2 / (i + 1)
            for i, word in enumerate(self._words)
        ]


class AnimalEmbeddings(WordEmbeddings):
    def __init__(self):
        super().__init__(
            words="""
            alli alpa amer amph ante ante antl appe aqua arct arma aust babo
            badg barr bask bear beav beet beha bird biso bite blac blue boar
            bobc brig buff bugl burr bush butt came cani capy cari carn cass
            cate cham chee chic chim chin chir clim coas coat cobr cock colo
            colo comm comp cour coyo crab cran croa croc crow crus wing wool
            cult cunn curi curl damb danc deer defe defe deme dese digg ding
            dise dist dive dolp dome dome donk dove drag drag duck ecos effo
            eigh elab eleg elev elon euca extr eyes falc famo famo fast fast
            feat feet ferr fier figh finc fish flam flig flig food fore foun
            fres frie frog gaze geck gees gent gill gira goat gori grac gras
            gras graz grou grou grou guin hams hard hawk hedg herb herd hero
            high hipp honk horn hors hove howl huma humm hump hunt hyen iden
            igua inde inse inte jack jagu jell jump jung kang koal komo lark
            larv lemu leop life lion liza lobs long loud loya mada magp mamm
            mana mari mars mass mati meat medi melo meta migr milk mimi moos
            mosq moth narw nati natu neck newt noct nort ocea octo ostr pack
            pain patt peac pest pinc pink play plum poll post pouc powe prec
            pred prey prid prim prob prot prow quac quil rais rapi reac rega
            rege regi rego rein rept resi rive roam rode sava scav seab seaf
            seas semi shar shed shel skil smal snak soci soft soli song song
            soun sout spec spee spik spor spot stag stic stin stin stor stre
            stre stre stro surv surv sust symb tail tall talo team teet tent
            term terr thou tiny tong toug tree agil tuft tund tusk umbr unic
            uniq vast vege veno vibr vita vora wadi wasp wate webb wetl wild
            ant bat bee cat cow dog eel elk emu fox pet pig""".split()
        )



# __init__.py

from .base import TraversalStrategy
from .eager import Eager
from .mmr import Mmr
from .scored import Scored

__all__ = [
    "Eager",
    "Scored",
    "Mmr",
    "TraversalStrategy",
]



from typing import Any

import pytest
from langchain_core.retrievers import BaseRetriever


@pytest.fixture(scope="function", params=["sync", "async"])
def invoker(request: pytest.FixtureRequest):
    async def sync_invoker(
        retriever: BaseRetriever, input: str, *args: Any, **kwargs: Any
    ):
        return retriever.invoke(input, *args, **kwargs)

    async def async_invoker(
        retriever: BaseRetriever, input: str, *args: Any, **kwargs: Any
    ):
        return await retriever.ainvoke(input, *args, **kwargs)

    if request.param == "sync":
        return sync_invoker
    elif request.param == "async":
        return async_invoker
    else:
        raise ValueError(f"Unexpected value '{request.param}'")



from typing import Iterable

from ..node import Node
from .base import TraversalStrategy


class Eager(TraversalStrategy):
    """Node selection that selects all nodes at each step."""

    _nodes: list[Node] = []

    def add_nodes(self, nodes: dict[str, Node]) -> None:
        self._nodes.extend(nodes.values())

    def select_nodes(self, *, limit: int) -> Iterable[Node]:
        nodes = self._nodes[:limit]
        self._nodes = []
        return nodes



import dataclasses
from functools import cached_property
from typing import Iterable

import numpy as np
from langchain_graph_retriever.utils.math import cosine_similarity
from numpy.typing import NDArray
from pydantic import Field

from ..node import Node
from .base import TraversalStrategy

NEG_INF = float("-inf")


def _emb_to_ndarray(embedding: list[float]) -> NDArray[np.float32]:
    emb_array = np.array(embedding, dtype=np.float32)
    if emb_array.ndim == 1:
        emb_array = np.expand_dims(emb_array, axis=0)
    return emb_array


@dataclasses.dataclass
class _MmrCandidate:
    node: Node
    similarity: float
    weighted_similarity: float
    weighted_redundancy: float
    score: float = dataclasses.field(init=False)

    def __post_init__(self) -> None:
        self.score = self.weighted_similarity - self.weighted_redundancy

    def update_redundancy(self, new_weighted_redundancy: float) -> None:
        if new_weighted_redundancy > self.weighted_redundancy:
            self.weighted_redundancy = new_weighted_redundancy
            self.score = self.weighted_similarity - self.weighted_redundancy


class Mmr(TraversalStrategy):
    """Helper for executing an MMR traversal query.
    Args:
        query_embedding: The embedding of the query to use for scoring.
        lambda_mult: Number between 0 and 1 that determines the degree
            of diversity among the results with 0 corresponding to maximum
            diversity and 1 to minimum diversity. Defaults to 0.5.
        score_threshold: Only documents with a score greater than or equal
            this threshold will be chosen. Defaults to -infinity.
    """

    lambda_mult: float = Field(default=0.5, ge=0.0, le=1.0)
    """Number between 0 and 1.

    Determines the degree of diversity among the results with 0 corresponding to
    maximum diversity and 1 to minimum diversity."""

    score_threshold: float = NEG_INF
    """Only documents with a score greater than or equal to this will be chosen."""

    _selected_ids: list[str] = []
    """List of selected IDs (in selection order)."""

    _candidate_id_to_index: dict[str, int] = {}
    """Dictionary of candidate IDs to indices in candidates and candidate_embeddings."""

    _candidates: list[_MmrCandidate] = []
    """List containing information about candidates.
    Same order as rows in `candidate_embeddings`.
    """

    _best_score: float = NEG_INF
    _best_id: str | None = None

    @cached_property
    def _nd_query_embedding(self) -> NDArray[np.float32]:
        assert (
            self.query_embedding
        ), "shouldn't access embedding / dimensions until initialized"
        return _emb_to_ndarray(self.query_embedding)

    @property
    def _dimensions(self) -> int:
        return self._nd_query_embedding.shape[1]

    @property
    def _lambda_mult_complement(self) -> float:
        return 1 - self.lambda_mult

    @cached_property
    def _selected_embeddings(self) -> NDArray[np.float32]:
        """(N, dim) ndarray with a row for each selected node."""
        return np.ndarray((self.k, self._dimensions), dtype=np.float32)

    @cached_property
    def _candidate_embeddings(self) -> NDArray[np.float32]:
        """(N, dim) ndarray with a row for each candidate."""
        return np.ndarray((0, self._dimensions), dtype=np.float32)

    def candidate_ids(self) -> Iterable[str]:
        """Return the IDs of the candidates."""
        return self._candidate_id_to_index.keys()

    def _already_selected_embeddings(self) -> NDArray[np.float32]:
        """Return the selected embeddings sliced to the already assigned values."""
        selected = len(self._selected_ids)
        return np.vsplit(self._selected_embeddings, [selected])[0]

    def _pop_candidate(
        self, candidate_id: str
    ) -> tuple[_MmrCandidate, NDArray[np.float32]]:
        """Pop the candidate with the given ID.
        Returns:
            The document, similarity score, and embedding of the candidate.
        """
        # Get the embedding for the id.
        index = self._candidate_id_to_index.pop(candidate_id)
        candidate = self._candidates[index]
        if candidate.node.id != candidate_id:
            msg = (
                "ID in self.candidate_id_to_index doesn't match the ID of the "
                "corresponding index in self.candidates"
            )
            raise ValueError(msg)
        embedding: NDArray[np.float32] = self._candidate_embeddings[index].copy()

        # Swap that index with the last index in the candidates and
        # candidate_embeddings.
        last_index = self._candidate_embeddings.shape[0] - 1

        if index == last_index:
            self._candidates.pop()
        else:
            self._candidate_embeddings[index] = self._candidate_embeddings[last_index]

            old_last = self._candidates.pop()
            self._candidates[index] = old_last
            self._candidate_id_to_index[old_last.node.id] = index

        self._candidate_embeddings = np.vsplit(
            self._candidate_embeddings, [last_index]
        )[0]

        return candidate, embedding

    def select_nodes(self, *, limit: int) -> Iterable[Node]:
        """Select and pop the best item being considered.
        Updates the consideration set based on it.
        Returns:
            A tuple containing the ID of the best item.
        """
        if limit == 0:
            return []
        if self._best_id is None or self._best_score < self.score_threshold:
            return []

        # Get the selection and remove from candidates.
        selected_id = self._best_id
        selected, selected_embedding = self._pop_candidate(selected_id)

        # Add the ID and embedding to the selected information.
        selection_index = len(self._selected_ids)
        self._selected_ids.append(selected_id)
        self._selected_embeddings[selection_index] = selected_embedding

        # Create the selected result node.
        selected_node = selected.node
        selected_node.extra_metadata = {
            "_similarity_score": selected.similarity,
            "_mmr_score": self._best_score,
        }

        # Reset the best score / best ID.
        self._best_score = NEG_INF
        self._best_id = None

        # Update the candidates redundancy, tracking the best node.
        if self._candidate_embeddings.shape[0] > 0:
            similarity = cosine_similarity(
                self._candidate_embeddings, np.expand_dims(selected_embedding, axis=0)
            )
            for index, candidate in enumerate(self._candidates):
                candidate.update_redundancy(similarity[index][0])
                if candidate.score > self._best_score:
                    self._best_score = candidate.score
                    self._best_id = candidate.node.id

        return [selected_node]

    def add_nodes(self, nodes: dict[str, Node]) -> None:
        """Add candidates to the consideration set."""
        # Determine the keys to actually include.
        # These are the candidates that aren't already selected
        # or under consideration.

        include_ids_set = set(nodes.keys())
        include_ids_set.difference_update(self._selected_ids)
        include_ids_set.difference_update(self._candidate_id_to_index.keys())
        include_ids = list(include_ids_set)

        # Now, build up a matrix of the remaining candidate embeddings.
        # And add them to the
        new_embeddings: NDArray[np.float32] = np.ndarray(
            (
                len(include_ids),
                self._dimensions,
            )
        )
        offset = self._candidate_embeddings.shape[0]
        for index, candidate_id in enumerate(include_ids):
            self._candidate_id_to_index[candidate_id] = offset + index
            new_embeddings[index] = nodes[candidate_id].embedding

        # Compute the similarity to the query.
        similarity = cosine_similarity(new_embeddings, self._nd_query_embedding)

        # Compute the distance metrics of all of pairs in the selected set with
        # the new candidates.
        redundancy = cosine_similarity(
            new_embeddings, self._already_selected_embeddings()
        )
        for index, candidate_id in enumerate(include_ids):
            max_redundancy = 0.0
            if redundancy.shape[0] > 0:
                max_redundancy = redundancy[index].max()
            candidate = _MmrCandidate(
                node=nodes[candidate_id],
                similarity=similarity[index][0],
                weighted_similarity=self.lambda_mult * similarity[index][0],
                weighted_redundancy=self._lambda_mult_complement * max_redundancy,
            )
            self._candidates.append(candidate)

            if candidate.score >= self._best_score:
                self._best_score = candidate.score
                self._best_id = candidate.node.id

        # Add the new embeddings to the candidate set.
        self._candidate_embeddings = np.vstack(
            (
                self._candidate_embeddings,
                new_embeddings,
            )
        )

    def finalize_nodes(self, nodes: Iterable[Node]) -> Iterable[Node]:
        return nodes



from typing import Any, List, Sequence

from langchain_core.documents import BaseDocumentTransformer, Document


class GLiNEREntityExtractor(BaseDocumentTransformer):
    """Add metadata to documents about named entities using `GLiNER`_.

    `GLiNER`_ is a Named Entity Recognition (NER) model capable of identifying any
    entity type using a bidirectional transformer encoder (BERT-like).

    Preliminaries
    -------------

    Install the ``gliner`` package.

    Note that ``bs4`` is also installed to support the WebBaseLoader in the example,
    but not needed by the GLiNEREntityExtractor itself.

    .. code-block:: bash

        pip install -q langchain_community bs4 gliner

    Example
    -----

    We load the ``state_of_the_union.txt`` file, chunk it, then for each chunk we
    add named entities to the metadata.

    .. code-block:: python

        from langchain_community.document_loaders import WebBaseLoader
        from langchain_community.document_transformers import GLiNEREntityExtractor
        from langchain_text_splitters import CharacterTextSplitter

        loader = WebBaseLoader(
            "https://raw.githubusercontent.com/hwchase17/chat-your-data/master/state_of_the_union.txt"
        )
        raw_documents = loader.load()

        text_splitter = CharacterTextSplitter(chunk_size=1000, chunk_overlap=0)
        documents = text_splitter.split_documents(raw_documents)

        extractor = GLiNEREntityExtractor(labels=["person", "topic"])
        documents = extractor.transform_documents(documents)

        print(documents[0].metadata)

    .. code-block:: output

        {'source': 'https://raw.githubusercontent.com/hwchase17/chat-your-data/master/state_of_the_union.txt', 'person': ['president zelenskyy', 'vladimir putin']}

    Args:
        labels: List of entity kinds to extract.
        batch_size: The number of documents to process in each batch (default ``8``)
        metadata_key_prefix: A prefix to add to metadata keys outputted by the extractor (default ``""``)
        model: The GLiNER model to use. (default ``urchade/gliner_mediumv2.1``)
    """  # noqa: E501

    def __init__(
        self,
        labels: List[str],
        *,
        batch_size: int = 8,
        metadata_key_prefix: str = "",
        model: str = "urchade/gliner_mediumv2.1",
    ):
        try:
            from gliner import GLiNER  # type: ignore

            self._model = GLiNER.from_pretrained(model)

        except ImportError:
            raise ImportError(
                "gliner is required for the GLiNEREntityExtractor. "
                "Please install it with `pip install gliner`."
            ) from None

        self._batch_size = batch_size
        self._labels = labels
        self.metadata_key_prefix = metadata_key_prefix

    def transform_documents(
        self, documents: Sequence[Document], **kwargs: Any
    ) -> Sequence[Document]:
        """Extracts named entities from documents using GLiNER.

        Args:
            documents: The sequence of documents to transform
            kwargs: Keyword arguments to pass to GLiNER. See: https://github.com/urchade/GLiNER/blob/v0.2.13/gliner/model.py#L419-L421
        """
        for i in range(0, len(documents), self._batch_size):
            batch = documents[i : i + self._batch_size]
            texts = [item.page_content for item in batch]
            extracted = self._model.batch_predict_entities(
                texts=texts, labels=self._labels, **kwargs
            )
            for i, entities in enumerate(extracted):
                labels = set()
                for entity in entities:
                    label = self.metadata_key_prefix + entity["label"]
                    labels.add(label)
                    batch[i].metadata.setdefault(label, set()).add(
                        entity["text"].lower()
                    )
                for label in labels:
                    batch[i].metadata[label] = list(batch[i].metadata[label])
        return documents



from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Optional,
    Sequence,
)

from langchain_core.documents import Document
from langchain_core.vectorstores import InMemoryVectorStore

from .base import METADATA_EMBEDDING_KEY, StoreAdapter

SENTINEL = object()


class InMemoryStoreAdapter(StoreAdapter[InMemoryVectorStore]):
    def __init__(
        self,
        vector_store: InMemoryVectorStore,
        *,
        support_normalized_metadata: bool = False,
    ):
        super().__init__(vector_store)
        self.support_normalized_metadata = support_normalized_metadata

    @property
    def supports_normalized_metadata(self) -> bool:
        return self.support_normalized_metadata

    def get(self, ids: Sequence[str], /, **kwargs) -> list[Document]:
        return self.vector_store.get_by_ids(ids)

    async def aget(self, ids: Sequence[str], /, **kwargs) -> list[Document]:
        return await self.vector_store.aget_by_ids(ids)

    def similarity_search_with_embedding_by_vector(
        self,
        embedding: list[float],
        k: int = 4,
        filter: Optional[Dict[str, str]] = None,
        **kwargs,
    ):
        results = self.vector_store._similarity_search_with_score_by_vector(
            embedding=embedding,
            k=k,
            filter=self._filter_method(filter),
        )
        return [
            Document(
                id=doc.id,
                page_content=doc.page_content,
                metadata={METADATA_EMBEDDING_KEY: embedding, **doc.metadata},
            )
            for doc, _score, embedding in results
        ]

    def _equals_or_contains(
        self,
        key: str,
        value: Any,
        metadata: dict[str, Any],
    ) -> bool:
        """Tests if the key->value exists or is contained in the metadata."""
        actual = metadata.get(key, SENTINEL)
        if actual == value:
            return True

        if (
            self.support_normalized_metadata
            and isinstance(actual, Iterable)
            and not isinstance(actual, (str, bytes))
            and value in actual
        ):
            return True

        return False

    def _filter_method(
        self, filter_dict: Optional[Dict[str, str]] = None
    ) -> Callable[[Document], bool]:
        if filter_dict is None:
            return lambda _doc: True

        def filter(doc: Document) -> bool:
            for key, value in filter_dict.items():
                if not self._equals_or_contains(key, value, doc.metadata):
                    return False
            return True

        return filter



import abc
from typing import Iterable

from pydantic import BaseModel

from ..node import Node


class TraversalStrategy(BaseModel, abc.ABC):
    """Interface for configuring node selection during the traversal."""

    k: int = 5
    """Number of nodes to retrieve during the traversal. Default 5."""

    start_k: int = 4
    """Number of initial documents to fetch via similarity.

    Will be added to the specified starting nodes, if any.
    """

    adjacent_k: int = 10
    """Number of adjacent Documents to fetch for each outgoing edge. Default 10.
    """

    max_depth: int | None = None
    """Maximum depth to retrieve. Default no limit."""

    query_embedding: list[float] = []
    """Query embedding."""

    @abc.abstractmethod
    def add_nodes(self, nodes: dict[str, Node]) -> None:
        """Add nodes to the set of available nodes."""
        ...

    @abc.abstractmethod
    def select_nodes(self, *, limit: int) -> Iterable[Node]:
        """Return the nodes to select at the next iteration.

        Iteration ends when this returns an empty list.
        """
        ...

    def finalize_nodes(self, nodes: Iterable[Node]) -> Iterable[Node]:
        """Finalize the selected nodes."""
        return nodes



from typing import (
    Any,
    Dict,
    List,
    Optional,
    Sequence,
)

from langchain_core.documents import Document

from .base import METADATA_EMBEDDING_KEY, StoreAdapter

try:
    from langchain_chroma import Chroma
except (ImportError, ModuleNotFoundError):
    msg = "please `pip install langchain-chroma`"
    raise ImportError(msg)


class ChromaStoreAdapter(StoreAdapter[Chroma]):
    def __init__(self, vector_store: Chroma):
        super().__init__(vector_store)

    @property
    def supports_normalized_metadata(self) -> bool:
        return False

    def similarity_search_with_embedding_by_vector(
        self,
        embedding: List[float],
        k: int = 4,
        filter: Optional[Dict[str, str]] = None,
        **kwargs: Any,
    ) -> List[Document]:
        """Returns docs (with embeddings) most similar to the query vector."""
        try:
            from chromadb.api.types import IncludeEnum
        except (ImportError, ModuleNotFoundError):
            msg = "please `pip install chromadb`"
            raise ImportError(msg)

        if k > self.vector_store._collection.count():
            k = self.vector_store._collection.count()

        results = self.vector_store._collection.query(
            query_embeddings=embedding,  # type: ignore
            n_results=k,
            where=filter,  # type: ignore
            include=[
                IncludeEnum.documents,
                IncludeEnum.metadatas,
                IncludeEnum.embeddings,
            ],
            **kwargs,
        )

        docs: list[Document] = []
        for result in zip(
            results["documents"][0],  # type: ignore
            results["metadatas"][0],  # type: ignore
            results["ids"][0],  # type: ignore
            results["embeddings"][0],  # type: ignore
        ):
            metadata = result[1] or {}
            metadata[METADATA_EMBEDDING_KEY] = result[3]
            docs.append(
                Document(
                    page_content=result[0],
                    metadata=metadata,
                    id=result[2],
                )
            )
        return docs

    def get(self, ids: Sequence[str], /, **kwargs: Any) -> list[Document]:
        """Get documents by id."""
        results = self.vector_store.get(ids=list(ids), **kwargs)
        return [
            Document(
                page_content=text,
                metadata=metadata,
                id=id,
            )
            for (text, metadata, id) in zip(
                results["documents"], results["metadatas"], results["ids"]
            )
        ]



import pytest
from langchain_graph_retriever.retrievers.edge import Edge
from langchain_graph_retriever.retrievers.edge_helper import EdgeHelper


def test_edge_helper_initialization():
    edge_helper = EdgeHelper(["a", ("b", "c"), "b"])
    assert edge_helper.edges == [("a", "a"), ("b", "c"), ("b", "b")]


def test_get_incoming_outgoing_normalized():
    edge_helper = EdgeHelper([("href", "url")], use_normalized_metadata=True)
    assert edge_helper.get_incoming_outgoing({"href": "a", "url": "b"}) == (
        {Edge("url", "b")},
        {Edge("url", "a")},
    )

    assert edge_helper.get_incoming_outgoing({"href": ["a", "c"], "url": "b"}) == (
        {Edge("url", "b")},
        {Edge("url", "a"), Edge("url", "c")},
    )

    assert edge_helper.get_incoming_outgoing(
        {"href": ["a", "c"], "url": ["b", "d"]}
    ) == ({Edge("url", "b"), Edge("url", "d")}, {Edge("url", "a"), Edge("url", "c")})


def test_get_incoming_outgoing_denormalized():
    edge_helper = EdgeHelper(
        [("href", "url")],
        use_normalized_metadata=False,
        # Use non-default values so we can verify the fields are used.
        denormalized_path_delimiter=":",
        denormalized_static_value=57,
    )
    assert edge_helper.get_incoming_outgoing({"href": "a", "url": "b"}) == (
        {Edge("url", "b")},
        {Edge("url", "a")},
    )

    assert edge_helper.get_incoming_outgoing(
        {"href:a": 57, "href:c": 57, "url": "b"}
    ) == ({Edge("url", "b")}, {Edge("url", "a"), Edge("url", "c")})

    assert edge_helper.get_incoming_outgoing(
        {
            "href:a": 57,
            "href:c": 57,
            "url:b": 57,
            "url:d": 57,
        }
    ) == ({Edge("url", "b"), Edge("url", "d")}, {Edge("url", "a"), Edge("url", "c")})


def test_get_incoming_outgoing_unsupported_values():
    edge_helper = EdgeHelper(
        [("href", "url")],
        use_normalized_metadata=True,
    )

    # Unsupported value
    with pytest.raises(ValueError, match=r"Unsupported value .* in 'href'"):
        edge_helper.get_incoming_outgoing({"href": None})

    # Unsupported item value
    with pytest.raises(ValueError, match=r"Unsupported item value .* in 'href'"):
        edge_helper.get_incoming_outgoing({"href": [None]})

    edge_helper = EdgeHelper(
        [("href", "url")],
        use_normalized_metadata=False,
        # Use non-default values so we can verify the fields are used.
        denormalized_path_delimiter=":",
        denormalized_static_value=57,
    )
    # Unsupported value
    with pytest.raises(ValueError, match=r"Unsupported value .* in 'href'"):
        edge_helper.get_incoming_outgoing({"href": None})

    # It is OK for the list to exist in the metadata, although we do issue a warning
    # for that case.
    with pytest.warns(UserWarning, match="Normalized value [[]'a', 'c'] in 'href'"):
        assert edge_helper.get_incoming_outgoing(
            {
                "href": ["a", "c"],
            }
        ) == (set(), set())


def test_get_metadata_filter_normalized():
    edge_helper = EdgeHelper([], use_normalized_metadata=True)

    assert edge_helper.get_metadata_filter(edge=Edge(key="boolean", value=True)) == {
        "boolean": True
    }

    assert edge_helper.get_metadata_filter(edge=Edge(key="incoming", value=4)) == {
        "incoming": 4
    }

    assert edge_helper.get_metadata_filter(edge=Edge(key="place", value="berlin")) == {
        "place": "berlin"
    }


def test_get_metadata_filter_denormalized() -> None:
    edge_helper = EdgeHelper([], use_normalized_metadata=False)

    assert edge_helper.get_metadata_filter(edge=Edge(key="boolean", value=True)) == {
        "boolean": True
    }

    assert edge_helper.get_metadata_filter(
        edge=Edge(key="incoming", value=4), denormalize_edge=True
    ) == {"incoming.4": "$"}

    assert edge_helper.get_metadata_filter(
        edge=Edge(key="place", value="berlin"), denormalize_edge=True
    ) == {"place.berlin": "$"}



import warnings
from typing import Any, Iterable

from .edge import Edge

BASIC_TYPES = (str, bool, int, float, complex, bytes)


# Sentinel object used with `dict.get(..., SENTINEL)` calls to distinguish
# between present but `None` (returns `None`) and absent (returns `SENTINEL`)
# elements.
SENTINEL = object()


class EdgeHelper:
    """Helper for extracting and encoding edges in metadata.

    Both incoming and outgoing edges are reported using the target name.
    This ensures that using them as keys allows equality matching.
    """

    def __init__(
        self,
        edges: list[str | tuple[str, str]],
        *,
        use_normalized_metadata: bool = False,
        denormalized_path_delimiter: str = ".",
        denormalized_static_value: Any = "$",
    ) -> None:
        self.use_normalized_metadata = use_normalized_metadata
        self.denormalized_path_delimiter = denormalized_path_delimiter
        self.denormalized_static_value = denormalized_static_value

        self.edges = []
        for edge in edges:
            if isinstance(edge, str):
                self.edges.append((edge, edge))
            elif (
                isinstance(edge, tuple)
                and len(edge) == 2
                and all(isinstance(item, str) for item in edge)
            ):
                self.edges.append((edge[0], edge[1]))
            else:
                raise ValueError(
                    "Invalid type for edge. must be 'str' or 'tuple[str,str]'"
                )

    def _edges_from_dict(
        self,
        metadata: dict[str, Any],
        *,
        warn_normalized: bool = False,
        incoming: bool = False,
    ) -> set[Edge]:
        """Extract edges from the metadata based on declared edges."""

        edges = set()
        for source_key, target_key in self.edges:
            if incoming:
                source_key = target_key

            value = metadata.get(source_key, SENTINEL)
            if isinstance(value, BASIC_TYPES):
                edges.add(Edge(target_key, value))
            elif isinstance(value, Iterable):
                # Note: `str` and `bytes` are in `BASIC_TYPES` so no need to
                # guard against.
                if warn_normalized:
                    warnings.warn(f"Normalized value {value} in '{source_key}'")
                else:
                    for item in value:
                        if isinstance(item, BASIC_TYPES):
                            edges.add(Edge(target_key, item))
                        else:
                            raise ValueError(
                                f"Unsupported item value {item} in '{source_key}'"
                            )
            elif value is not SENTINEL:
                raise ValueError(f"Unsupported value {value} in '{source_key}'")
        return edges

    def _normalize_metadata(
        self, denormalized_metadata: dict[str, Any]
    ) -> dict[str, Any]:
        normalized: dict[str, Any] = {}
        for key, value in denormalized_metadata.items():
            try:
                if value != self.denormalized_static_value:
                    continue
            except (TypeError, ValueError):
                # Skip items that can't be compared
                continue

            split = key.split(self.denormalized_path_delimiter, 2)
            if len(split) == 2 and len(split[1]) > 0:
                normalized.setdefault(split[0], set()).add(split[1])
        return normalized

    def get_incoming_outgoing(
        self, metadata: dict[str, Any]
    ) -> tuple[set[Edge], set[Edge]]:
        warn_normalized = not self.use_normalized_metadata
        outgoing_edges = self._edges_from_dict(
            metadata, warn_normalized=warn_normalized
        )
        incoming_edges = self._edges_from_dict(
            metadata, incoming=True, warn_normalized=warn_normalized
        )

        if not self.use_normalized_metadata:
            normalized = self._normalize_metadata(metadata)

            outgoing_edges.update(self._edges_from_dict(normalized))
            incoming_edges.update(self._edges_from_dict(normalized, incoming=True))

        return (incoming_edges, outgoing_edges)

    def get_metadata_filter(
        self,
        base_filter: dict[str, Any] | None = None,
        edge: Edge | None = None,
        denormalize_edge: bool = False,
    ) -> dict[str, Any]:
        """Builds a metadata filter to search for documents

        Args:
            base_filter: Any metadata that should be used for hybrid search
            edge: An optional outgoing edge to add to the search
        """
        metadata_filter = {**(base_filter or {})}
        if edge is None:
            metadata_filter
        elif denormalize_edge:
            metadata_filter[
                f"{edge.key}{self.denormalized_path_delimiter}{edge.value}"
            ] = self.denormalized_static_value
        else:
            metadata_filter[edge.key] = edge.value
        return metadata_filter



from typing import (
    Any,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
)

try:
    from langchain_astradb import AstraDBVectorStore
except (ImportError, ModuleNotFoundError):
    raise ImportError("please `pip install langchain-astradb`")
from langchain_core.documents import Document

from .base import METADATA_EMBEDDING_KEY, StoreAdapter


class AstraStoreAdapter(StoreAdapter[AstraDBVectorStore]):
    def __init__(self, vector_store: AstraDBVectorStore):
        self.vector_store = vector_store

    @property
    def supports_normalized_metadata(self) -> bool:
        return True

    def _build_docs(
        self, docs_with_embeddings: list[tuple[Document, list[float]]]
    ) -> List[Document]:
        docs: List[Document] = []
        for doc, embedding in docs_with_embeddings:
            doc.metadata[METADATA_EMBEDDING_KEY] = embedding
            docs.append(doc)
        return docs

    def similarity_search_with_embedding(
        self,
        query: str,
        k: int = 4,
        filter: Optional[Dict[str, str]] = None,
        **kwargs: Any,
    ) -> Tuple[List[float], List[Document]]:
        """Returns docs (with embeddings) most similar to the query."""
        query_embedding, docs_with_embeddings = (
            self.vector_store.similarity_search_with_embedding(
                query=query,
                k=k,
                filter=filter,
                **kwargs,
            )
        )
        return query_embedding, self._build_docs(
            docs_with_embeddings=docs_with_embeddings
        )

    async def asimilarity_search_with_embedding(
        self,
        query: str,
        k: int = 4,
        filter: Optional[Dict[str, str]] = None,
        **kwargs: Any,
    ) -> Tuple[List[float], List[Document]]:
        """Returns docs (with embeddings) most similar to the query."""
        (
            query_embedding,
            docs_with_embeddings,
        ) = await self.vector_store.asimilarity_search_with_embedding(
            query=query,
            k=k,
            filter=filter,
            **kwargs,
        )
        return query_embedding, self._build_docs(
            docs_with_embeddings=docs_with_embeddings
        )

    def similarity_search_with_embedding_by_vector(
        self,
        embedding: List[float],
        k: int = 4,
        filter: Optional[Dict[str, str]] = None,
        **kwargs: Any,
    ) -> List[Document]:
        """Returns docs (with embeddings) most similar to the query vector."""
        docs_with_embeddings = (
            self.vector_store.similarity_search_with_embedding_by_vector(
                embedding=embedding,
                k=k,
                filter=filter,
                **kwargs,
            )
        )
        return self._build_docs(docs_with_embeddings=docs_with_embeddings)

    async def asimilarity_search_with_embedding_by_vector(
        self,
        embedding: List[float],
        k: int = 4,
        filter: Optional[Dict[str, str]] = None,
        **kwargs: Any,
    ) -> List[Document]:
        """Returns docs (with embeddings) most similar to the query vector."""
        docs_with_embeddings = (
            await self.vector_store.asimilarity_search_with_embedding_by_vector(
                embedding=embedding,
                k=k,
                filter=filter,
                **kwargs,
            )
        )
        return self._build_docs(docs_with_embeddings=docs_with_embeddings)

    def get(self, ids: Sequence[str], /, **kwargs: Any) -> list[Document]:
        """Get documents by id."""
        docs: list[Document] = []
        for id in ids:
            doc = self.vector_store.get_by_document_id(id, **kwargs)
            if doc is not None:
                docs.append(doc)
        return docs

    async def aget(self, ids: Sequence[str], /, **kwargs: Any) -> list[Document]:
        """Get documents by id."""
        docs: list[Document] = []
        for id in ids:
            doc = await self.vector_store.aget_by_document_id(id, **kwargs)
            if doc is not None:
                docs.append(doc)
        return docs



from .simple_embeddings import (
    Angular2DEmbeddings,
    AnimalEmbeddings,
    EarthEmbeddings,
    ParserEmbeddings,
)

__all__ = [
    "Angular2DEmbeddings",
    "AnimalEmbeddings",
    "EarthEmbeddings",
    "ParserEmbeddings",
]



from typing import Any, Sequence

from langchain_core.documents import BaseDocumentTransformer, Document


class MetadataDenormalizer(BaseDocumentTransformer):
    """Denormalizes sequence-based metadata fields

    Certain vector stores do not support storing or searching on metadata fields
    with sequence-based values. This transformer converts sequence-based fields
    into simple metadata values.

    Example
    -----

    .. code-block:: python

        from langchain_core.documents import Document
        from langchain_community.document_transformers.metadata_denormalizer import (
            MetadataDenormalizer,
        )

        doc = Document(
            page_content="test",
            metadata={"place": ["berlin", "paris"], "topic": ["weather"]},
        )

        de_normalizer = MetadataDenormalizer()

        docs = de_normalizer.transform_documents([doc])

        print(docs[0].metadata)


    .. code-block:: output

        {'place.berlin': True, 'place.paris': True, 'topic.weather': True}

    Args:
        keys: A set of metadata keys to denormalize. If empty, all
            sequence-based fields will be denormalized.
        path_delimiter: The path delimiter to use when building denormalized keys.
        static_value: The value to set on each denormalized key.
    """  # noqa: E501

    def __init__(
        self,
        *,
        keys: set[str] = set(),
        path_delimiter: str = ".",
        static_value: Any = "$",
    ):
        self.keys = keys
        self.path_delimiter = path_delimiter
        self.static_value = static_value

    def transform_documents(
        self, documents: Sequence[Document], **kwargs: Any
    ) -> Sequence[Document]:
        """Denormalizes sequence-based metadata fields."""
        transformed_docs = []
        for document in documents:
            new_doc = Document(id=document.id, page_content=document.page_content)
            for key, value in document.metadata.items():
                is_normalized = isinstance(value, Sequence) and not isinstance(
                    value, (str, bytes)
                )
                should_denormalize = (not self.keys) or (key in self.keys)
                if is_normalized and should_denormalize:
                    for item in value:
                        new_doc.metadata[f"{key}{self.path_delimiter}{item}"] = (
                            self.static_value
                        )
                else:
                    new_doc.metadata[key] = value
            transformed_docs.append(new_doc)

        return transformed_docs



from typing import Iterable

from langchain_core.documents import Document


def sorted_doc_ids(docs: Iterable[Document]) -> list[str]:
    return sorted([doc.id for doc in docs if doc.id is not None])


def assert_document_format(doc: Document) -> None:
    assert doc.id is not None
    assert doc.page_content is not None
    assert doc.metadata is not None
    assert "__embedding" not in doc.metadata



from tests.integration_tests.invoker import invoker
from tests.integration_tests.retrievers.animal_docs import animal_docs, animal_store
from tests.integration_tests.retrievers.earth_docs import (
    earth_docs,
    earth_store,
)
from tests.integration_tests.retrievers.parser_docs import (
    graph_vector_store_docs,
    parser_store,
)

# Imports for definitions.
from tests.integration_tests.stores import (
    enabled_stores,
    store_factory,
    store_param,
)

# Mark these imports as used so they don't removed.
# They need to be imported here so the fixtures are available.
_ = (
    store_factory,
    store_param,
    enabled_stores,
    earth_docs,
    earth_store,
    animal_docs,
    animal_store,
    graph_vector_store_docs,
    parser_store,
    invoker,
)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, List, Optional, Sequence, Set
from urllib.parse import urldefrag, urljoin, urlparse

from langchain_core._api import beta
from langchain_core.documents import BaseDocumentTransformer, Document

if TYPE_CHECKING:
    from bs4 import BeautifulSoup  # type: ignore
    from bs4.element import Tag  # type: ignore


@beta()
class HtmlHyperlinkExtractor(BaseDocumentTransformer):
    """Extract hyperlinks from HTML content.

    Expects each document to contain its URL in its metadata.

    Example::

        extractor = HtmlHyperlinkExtractor()
        results = extractor.extract_one(HtmlInput(html, url))

    .. seealso::

        - :mod:`How to use a graph vector store <langchain_community.graph_vectorstores>`
        - :class:`How to create links between documents <langchain_community.graph_vectorstores.links.Link>`

    How to link Documents on hyperlinks in HTML
    ===========================================

    Preliminaries
    -------------

    Install the ``beautifulsoup4`` package:

    .. code-block:: bash

        pip install -q langchain_community beautifulsoup4

    Usage
    -----

    For this example, we'll scrape 2 HTML pages that have an hyperlink from one
    page to the other using an ``AsyncHtmlLoader``.
    Then we use the ``HtmlLinkExtractor`` to create the links in the documents.

    Using extract_one()
    ^^^^^^^^^^^^^^^^^^^

    We can use :meth:`extract_one` on a document to get the links and add the links
    to the document metadata with
    :meth:`~langchain_community.graph_vectorstores.links.add_links`::

        from langchain_community.document_loaders import AsyncHtmlLoader
        from langchain_community.graph_vectorstores.extractors import (
            HtmlInput,
            HtmlLinkExtractor,
        )
        from langchain_community.graph_vectorstores.links import add_links
        from langchain_core.documents import Document

        loader = AsyncHtmlLoader(
            [
                "https://python.langchain.com/docs/integrations/providers/astradb/",
                "https://docs.datastax.com/en/astra/home/astra.html",
            ]
        )

        documents = loader.load()

        html_extractor = HtmlLinkExtractor()

        for doc in documents:
            links = html_extractor.extract_one(HtmlInput(doc.page_content, url))
            add_links(doc, links)

        documents[0].metadata["links"][:5]

    .. code-block:: output

        [Link(kind='hyperlink', direction='out', tag='https://python.langchain.com/docs/integrations/providers/spreedly/'),
            Link(kind='hyperlink', direction='out', tag='https://python.langchain.com/docs/integrations/providers/nvidia/'),
            Link(kind='hyperlink', direction='out', tag='https://python.langchain.com/docs/integrations/providers/ray_serve/'),
            Link(kind='hyperlink', direction='out', tag='https://python.langchain.com/docs/integrations/providers/bageldb/'),
            Link(kind='hyperlink', direction='out', tag='https://python.langchain.com/docs/introduction/')]

    Using as_document_extractor()
    ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    If you use a document loader that returns the raw HTML and that sets the source
    key in the document metadata such as ``AsyncHtmlLoader``,
    you can simplify by using :meth:`as_document_extractor` that takes directly a
    ``Document`` as input::

        from langchain_community.document_loaders import AsyncHtmlLoader
        from langchain_community.graph_vectorstores.extractors import HtmlLinkExtractor
        from langchain_community.graph_vectorstores.links import add_links

        loader = AsyncHtmlLoader(
            [
                "https://python.langchain.com/docs/integrations/providers/astradb/",
                "https://docs.datastax.com/en/astra/home/astra.html",
            ]
        )
        documents = loader.load()
        html_extractor = HtmlLinkExtractor().as_document_extractor()

        for document in documents:
            links = html_extractor.extract_one(document)
            add_links(document, links)

        documents[0].metadata["links"][:5]

    .. code-block:: output

        [Link(kind='hyperlink', direction='out', tag='https://python.langchain.com/docs/integrations/providers/spreedly/'),
            Link(kind='hyperlink', direction='out', tag='https://python.langchain.com/docs/integrations/providers/nvidia/'),
            Link(kind='hyperlink', direction='out', tag='https://python.langchain.com/docs/integrations/providers/ray_serve/'),
            Link(kind='hyperlink', direction='out', tag='https://python.langchain.com/docs/integrations/providers/bageldb/'),
            Link(kind='hyperlink', direction='out', tag='https://python.langchain.com/docs/introduction/')]

    Using LinkExtractorTransformer
    ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    Using the :class:`~langchain_community.graph_vectorstores.extractors.link_extractor_transformer.LinkExtractorTransformer`,
    we can simplify the link extraction::

        from langchain_community.document_loaders import AsyncHtmlLoader
        from langchain_community.graph_vectorstores.extractors import (
            HtmlLinkExtractor,
            LinkExtractorTransformer,
        )
        from langchain_community.graph_vectorstores.links import add_links

        loader = AsyncHtmlLoader(
            [
                "https://python.langchain.com/docs/integrations/providers/astradb/",
                "https://docs.datastax.com/en/astra/home/astra.html",
            ]
        )

        documents = loader.load()
        transformer = LinkExtractorTransformer(
            [HtmlLinkExtractor().as_document_extractor()]
        )
        documents = transformer.transform_documents(documents)

        documents[0].metadata["links"][:5]

    .. code-block:: output

        [Link(kind='hyperlink', direction='out', tag='https://python.langchain.com/docs/integrations/providers/spreedly/'),
            Link(kind='hyperlink', direction='out', tag='https://python.langchain.com/docs/integrations/providers/nvidia/'),
            Link(kind='hyperlink', direction='out', tag='https://python.langchain.com/docs/integrations/providers/ray_serve/'),
            Link(kind='hyperlink', direction='out', tag='https://python.langchain.com/docs/integrations/providers/bageldb/'),
            Link(kind='hyperlink', direction='out', tag='https://python.langchain.com/docs/introduction/')]

    We can check that there is a link from the first document to the second::

        for doc_to in documents:
            for link_to in doc_to.metadata["links"]:
                if link_to.direction == "in":
                    for doc_from in documents:
                        for link_from in doc_from.metadata["links"]:
                            if (
                                link_to.direction == "in"
                                and link_from.direction == "out"
                                and link_to.tag == link_from.tag
                            ):
                                print(
                                    f"Found link from {doc_from.metadata['source']} to {doc_to.metadata['source']}."
                                )

    .. code-block:: output

        Found link from https://python.langchain.com/docs/integrations/providers/astradb/ to https://docs.datastax.com/en/astra/home/astra.html.

    The documents with URL links can then be added to a :class:`~langchain_community.graph_vectorstores.base.GraphVectorStore`::

        from langchain_community.graph_vectorstores import CassandraGraphVectorStore

        store = CassandraGraphVectorStore.from_documents(documents=documents, embedding=...)

    Args:
        kind: The kind of edge to extract. Defaults to ``hyperlink``.
        drop_fragments: Whether fragments in URLs and links should be
            dropped. Defaults to ``True``.
    """  # noqa: E501

    def __init__(
        self,
        *,
        url_metadata_key: str = "url",
        metadata_key: str = "hyperlink",
        drop_fragments: bool = True,
    ):
        try:
            from bs4 import BeautifulSoup  # noqa:F401
        except ImportError as e:
            raise ImportError(
                "BeautifulSoup4 is required for HtmlHyperlinkExtractor. "
                "Please install it with `pip install beautifulsoup4`."
            ) from e

        self._url_metadata_key = url_metadata_key
        self._metadata_key = metadata_key
        self._drop_fragments = drop_fragments

    @staticmethod
    def _parse_url(
        link: Tag, page_url: str, drop_fragments: bool = True
    ) -> Optional[str]:
        href = link.get("href")
        if href is None:
            return None
        url = urlparse(href)
        if url.scheme not in ["http", "https", ""]:
            return None

        # Join the HREF with the page_url to convert relative paths to absolute.
        url = str(urljoin(page_url, href))

        # Fragments would be useful if we chunked a page based on section.
        # Then, each chunk would have a different URL based on the fragment.
        # Since we aren't doing that yet, they just "break" links. So, drop
        # the fragment.
        if drop_fragments:
            return urldefrag(url).url
        return url

    @staticmethod
    def _parse_urls(
        soup: BeautifulSoup, page_url: str, drop_fragments: bool = True
    ) -> List[str]:
        soup_links: List[Tag] = soup.find_all("a")
        urls: Set[str] = set()

        for link in soup_links:
            parsed_url = HtmlHyperlinkExtractor._parse_url(
                link, page_url=page_url, drop_fragments=drop_fragments
            )
            # Remove self links and entries for any 'a' tag that failed to parse
            # (didn't have href, or invalid domain, etc.)
            if parsed_url and parsed_url != page_url:
                urls.add(parsed_url)

        return list(urls)

    def transform_documents(
        self, documents: Sequence[Document], **kwargs: Any
    ) -> Sequence[Document]:
        """Extracts hyperlinks from html documents using BeautifulSoup4"""
        for document in documents:
            if self._url_metadata_key not in document.metadata:
                msg = (
                    f"html document url not found in metadata at "
                    f"{self._url_metadata_key} for document id: {document.id}"
                )
                raise ValueError(msg)

            page_url = document.metadata[self._url_metadata_key]
            if self._drop_fragments:
                page_url = urldefrag(page_url).url

            soup = BeautifulSoup(document.page_content, "html.parser", **kwargs)

            document.metadata[self._metadata_key] = self._parse_urls(
                soup=soup, page_url=page_url, drop_fragments=self._drop_fragments
            )
        return documents



import abc
from typing import Callable, Generic, TypeVar

import pytest
from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings
from langchain_core.vectorstores import VectorStore
from langchain_graph_retriever.document_transformers.metadata_denormalizer import (
    MetadataDenormalizer,
)
from langchain_graph_retriever.retrievers.store_adapters import StoreAdapter

ALL_STORES = ["mem_norm", "mem", "astra", "cassandra", "chroma", "opensearch"]
TESTCONTAINER_STORES = ["cassandra", "opensearch"]


@pytest.fixture(scope="session")
def enabled_stores(request: pytest.FixtureRequest) -> set[str]:
    # TODO: Use StrEnum?
    stores = request.config.getoption("--stores")

    if stores and "all" in stores:
        return set(ALL_STORES)
    elif stores:
        return set(stores)
    else:
        return {"mem_norm", "mem"}


def use_testcontainer(request: pytest.FixtureRequest, store: str) -> bool:
    testcontainers = request.config.getoption("--testcontainer")
    if testcontainers and "none" in testcontainers:
        return False
    elif testcontainers:
        return store in testcontainers
    else:
        return True


@pytest.fixture(scope="session", params=ALL_STORES)
def store_param(request: pytest.FixtureRequest, enabled_stores: set[str]) -> str:
    store: str = request.param
    if store not in enabled_stores:
        pytest.skip(f"'{store}' is not enabled")
    return store


T = TypeVar("T", bound=VectorStore)


class StoreFactory(abc.ABC, Generic[T]):
    def __init__(
        self,
        create_store: Callable[[str, list[Document], Embeddings], T],
        create_adapter: Callable[[T], StoreAdapter],
        teardown: Callable[[T], None] | None = None,
    ):
        self._create_store = create_store
        self._create_adapter = create_adapter
        self._teardown = teardown
        self._index = 0

    def create(
        self,
        request: pytest.FixtureRequest,
        embedding: Embeddings,
        docs: list[Document],
    ) -> StoreAdapter:
        name = f"test_{self._index}"
        self._index += 1
        store = self._create_store(name, docs, embedding)

        if self._teardown is not None:
            # make a local copy of the non-None teardown. This makes `mypy` happy.
            # Otherwise, it (correctly) recognizes that `self._teardown` could be not
            # `None` and `None` later (when the finalizer is called)
            teardown = self._teardown
            request.addfinalizer(lambda: teardown(store))

        return self._create_adapter(store)


def _cassandra_store_factory(request: pytest.FixtureRequest):
    import os

    from cassandra.cluster import Cluster  # type: ignore
    from langchain_community.vectorstores.cassandra import Cassandra
    from langchain_graph_retriever.retrievers.store_adapters.cassandra import (
        CassandraStoreAdapter,
    )

    if use_testcontainer(request, "cassandra"):
        from testcontainers.cassandra import CassandraContainer  # type: ignore

        container = CassandraContainer()
        container.start()
        request.addfinalizer(lambda: container.stop())
        contact_points = container.get_contact_points()
    elif "CASSANDRA_CONTACT_POINTS" in os.environ:
        contact_points = [
            cp.strip()
            for cp in os.environ["CASSANDRA_CONTACT_POINTS"].split(",")
            if cp.strip()
        ]
    else:
        contact_points = None

    cluster = Cluster(contact_points)
    session = cluster.connect()

    KEYSPACE = "graph_test_keyspace"
    session.execute(
        (
            f"CREATE KEYSPACE IF NOT EXISTS {KEYSPACE}"
            " WITH replication = "
            "{'class': 'SimpleStrategy', 'replication_factor': 1}"
        )
    )

    request.addfinalizer(lambda: cluster.shutdown())

    def create_cassandra(
        name: str, docs: list[Document], embedding: Embeddings
    ) -> Cassandra:
        session = cluster.connect()
        session.execute(f"DROP TABLE IF EXISTS {KEYSPACE}.{name}")

        store = Cassandra(
            embedding=embedding,
            session=session,
            keyspace=KEYSPACE,
            table_name=name,
        )
        docs = list(MetadataDenormalizer().transform_documents(docs))
        store.add_documents(docs)
        return store

    def teardown_cassandra(cassandra: Cassandra):
        assert cassandra.session is not None
        cassandra.session.shutdown()

    return StoreFactory[Cassandra](
        create_store=create_cassandra,
        create_adapter=CassandraStoreAdapter,
        teardown=teardown_cassandra,
    )


def _opensearch_store_factory(request: pytest.FixtureRequest):
    from langchain_community.vectorstores import OpenSearchVectorSearch
    from langchain_graph_retriever.retrievers.store_adapters.open_search import (
        OpenSearchStoreAdapter,
    )

    if use_testcontainer(request, "opensearch"):
        from testcontainers.opensearch import OpenSearchContainer  # type: ignore

        # If the admin password doesn't pass the length and regex requirements
        # starting the container will hang (`docker ps <container_id>` to debug).
        container = OpenSearchContainer(
            image="opensearchproject/opensearch:2.18.0",
            initial_admin_password="SomeRandomP4ssword",
        )
        container.start()
        request.addfinalizer(lambda: container.stop())

        config = container.get_config()
        opensearch_url = f"http://{config['host']}:{config['port']}"
        kwargs = {"http_auth": (config["username"], config["password"])}
    else:
        opensearch_url = "http://localhost:9200"
        kwargs = {}

    def create_open_search(
        name: str, docs: list[Document], embedding: Embeddings
    ) -> OpenSearchVectorSearch:
        store = OpenSearchVectorSearch(
            opensearch_url=opensearch_url,
            index_name=name,
            embedding_function=embedding,
            engine="faiss",
            **kwargs,
        )
        store.add_documents(docs)
        return store

    def teardown_open_search(store: OpenSearchVectorSearch) -> None:
        if store.index_exists():
            store.delete_index()

    return StoreFactory[OpenSearchVectorSearch](
        create_store=create_open_search,
        create_adapter=OpenSearchStoreAdapter,
        teardown=teardown_open_search,
    )


def _astra_store_factory(_request: pytest.FixtureRequest) -> StoreFactory:
    import os

    from astrapy.authentication import StaticTokenProvider
    from dotenv import load_dotenv
    from langchain_astradb import AstraDBVectorStore
    from langchain_graph_retriever.retrievers.store_adapters.astra import (
        AstraStoreAdapter,
    )

    load_dotenv()

    token = StaticTokenProvider(os.environ["ASTRA_DB_APPLICATION_TOKEN"])
    keyspace = os.environ.get("ASTRA_DB_KEYSPACE", "default_keyspace")
    api_endpoint = os.environ["ASTRA_DB_API_ENDPOINT"]

    def create_astra(
        name: str, docs: list[Document], embedding: Embeddings
    ) -> AstraDBVectorStore:
        try:
            from langchain_astradb import AstraDBVectorStore

            store = AstraDBVectorStore(
                embedding=embedding,
                collection_name=name,
                namespace=keyspace,
                token=token,
                api_endpoint=api_endpoint,
            )
            store.add_documents(docs)
            return store

        except (ImportError, ModuleNotFoundError):
            msg = (
                "to test graph-traversal with AstraDB, please"
                " install langchain-astradb and python-dotenv"
            )
            raise ImportError(msg)

    def teardown_astra(store: AstraDBVectorStore):
        store.delete_collection()

    return StoreFactory[AstraDBVectorStore](
        create_store=create_astra,
        create_adapter=AstraStoreAdapter,
        teardown=teardown_astra,
    )


def _in_memory_store_factory(
    _request: pytest.FixtureRequest, support_normalized_metadata: bool
) -> StoreFactory:
    from langchain_core.vectorstores import InMemoryVectorStore
    from langchain_graph_retriever.retrievers.store_adapters.in_memory import (
        InMemoryStoreAdapter,
    )

    def create_in_memory(
        _name: str, docs: list[Document], emb: Embeddings
    ) -> InMemoryVectorStore:
        if not support_normalized_metadata:
            docs = list(MetadataDenormalizer().transform_documents(docs))
        return InMemoryVectorStore.from_documents(docs, emb)

    return StoreFactory[InMemoryVectorStore](
        create_store=create_in_memory,
        create_adapter=lambda store: InMemoryStoreAdapter(
            store, support_normalized_metadata=support_normalized_metadata
        ),
    )


def _chroma_store_factory(_request: pytest.FixtureRequest) -> StoreFactory:
    from langchain_chroma.vectorstores import Chroma
    from langchain_graph_retriever.retrievers.store_adapters.chroma import (
        ChromaStoreAdapter,
    )

    def create_chroma(name: str, docs: list[Document], emb: Embeddings) -> Chroma:
        docs = list(MetadataDenormalizer().transform_documents(docs))
        return Chroma.from_documents(docs, emb, collection_name=name)

    return StoreFactory[Chroma](
        create_store=create_chroma,
        create_adapter=ChromaStoreAdapter,
        teardown=lambda store: store.delete_collection(),
    )


@pytest.fixture(scope="session")
def store_factory(store_param: str, request: pytest.FixtureRequest) -> StoreFactory:
    if store_param == "mem_norm":
        return _in_memory_store_factory(request, support_normalized_metadata=True)
    elif store_param == "mem":
        return _in_memory_store_factory(request, support_normalized_metadata=False)
    elif store_param == "chroma":
        return _chroma_store_factory(request)
    elif store_param == "astra":
        return _astra_store_factory(request)
    elif store_param == "cassandra":
        return _cassandra_store_factory(request)
    elif store_param == "opensearch":
        return _opensearch_store_factory(request)
    else:
        pytest.fail(f"Unsupported store: {store_param}")



import json
import os

import pytest
from langchain_core.documents import Document
from tests.embeddings import AnimalEmbeddings
from tests.integration_tests.stores import StoreAdapter, StoreFactory


@pytest.fixture(scope="session")
def animal_docs() -> list[Document]:
    documents = []

    path = os.path.abspath(
        os.path.join(os.path.dirname(__file__), "../../../../../data/animals.jsonl")
    )
    with open(path, "r") as file:
        for line in file:
            data = json.loads(line.strip())
            documents.append(
                Document(
                    id=data["id"],
                    page_content=data["text"],
                    metadata=data["metadata"],
                )
            )

    return documents


@pytest.fixture(scope="session")
def animal_store(
    request: pytest.FixtureRequest,
    store_factory: StoreFactory,
    animal_docs: list[Document],
) -> StoreAdapter:
    return store_factory.create(request, AnimalEmbeddings(), animal_docs)


ANIMALS_QUERY: str = "small agile mammal"
ANIMALS_DEPTH_0_EXPECTED: list[str] = ["fox", "mongoose"]



import pytest
from langchain_core.documents import Document
from langchain_graph_retriever.retrievers.store_adapters import StoreAdapter
from tests.embeddings.simple_embeddings import EarthEmbeddings
from tests.integration_tests.stores import StoreFactory


@pytest.fixture(scope="session")
def earth_docs() -> list[Document]:
    """This is a set of Documents to pre-populate a graph vector store."""

    greetings = Document(
        id="greetings",
        page_content="Typical Greetings",
        metadata={
            "incoming": "parent",
        },
    )

    doc1 = Document(
        id="doc1",
        page_content="Hello World",
        metadata={"outgoing": "parent", "keywords": ["greeting", "world"]},
    )

    doc2 = Document(
        id="doc2",
        page_content="Hello Earth",
        metadata={"outgoing": "parent", "keywords": ["greeting", "earth"]},
    )
    return [greetings, doc1, doc2]


@pytest.fixture(scope="session")
def earth_store(
    request: pytest.FixtureRequest,
    store_factory: StoreFactory,
    earth_docs: list[Document],
) -> StoreAdapter:
    return store_factory.create(request, EarthEmbeddings(), earth_docs)



from typing import Any, Sequence

from langchain_core.documents import BaseDocumentTransformer, Document


class HierarchyLinkExtractor(BaseDocumentTransformer):
    def __init__(
        self,
        *,
        path_metadata_key: str = "path",
        path_delimiter: str = "\\",
        parent_links: bool = True,
        child_links: bool = False,
        sibling_links: bool = False,
    ):
        """Extract links from a document hierarchy.

        Example:

            .. code-block:: python

                # Given three paths (in this case, within the "Root" document):
                h1 = ["Root", "H1"]
                h1a = ["Root", "H1", "a"]
                h1b = ["Root", "H1", "b"]

                # Parent links `h1a` and `h1b` to `h1`.
                # Child links `h1` to `h1a` and `h1b`.
                # Sibling links `h1a` and `h1b` together (both directions).

        Example use with documents:
            .. code_block: python
                transformer = LinkExtractorTransformer([
                    HierarchyLinkExtractor().as_document_extractor(
                        # Assumes the "path" to each document is in the metadata.
                        # Could split strings, etc.
                        lambda doc: doc.metadata.get("path", [])
                    )
                ])
                linked = transformer.transform_documents(docs)

        Args:
            kind: Kind of links to produce with this extractor.
            parent_links: Link from a section to its parent.
            child_links: Link from a section to its children.
            sibling_links: Link from a section to other sections with the same parent.
        """
        self._path_metadata_key = path_metadata_key
        self._path_delimiter = path_delimiter
        self._parent_links = parent_links
        self._child_links = child_links
        self._sibling_links = sibling_links

    def transform_documents(
        self, documents: Sequence[Document], **kwargs: Any
    ) -> Sequence[Document]:
        """Extracts hyperlinks from html documents using BeautifulSoup4"""
        for document in documents:
            if self._path_metadata_key not in document.metadata:
                msg = (
                    f"path not found in metadata at {self._path_metadata_key}",
                    f" for document id: {document.id}",
                )
                raise ValueError(msg)

            path: str = document.metadata[self._path_metadata_key]
            if self._parent_links:
                # This is linked from everything with this parent path.
                document.metadata["parent_in"] = path
            if self._child_links:
                # This is linked to every child with this as it's "parent" path.
                document.metadata["child_out"] = path

            path_parts = path.split(self._path_delimiter)
            if len(path_parts) >= 1:
                parent_path = self._path_delimiter.join(path_parts[0:-1])
                if self._parent_links and len(path_parts) > 1:
                    # This is linked to the nodes with the given parent path.
                    document.metadata["parent_out"] = parent_path
                if self._child_links and len(path_parts) > 1:
                    # This is linked from every node with the given parent path.
                    document.metadata["child_in"] = parent_path
                if self._sibling_links:
                    # This is a sibling of everything with the same parent.
                    document.metadata["sibling"] = parent_path
        return documents



import pytest
from langchain_core.documents import Document
from tests.embeddings.simple_embeddings import ParserEmbeddings
from tests.integration_tests.stores import StoreAdapter, StoreFactory


@pytest.fixture(scope="session")
def graph_vector_store_docs() -> list[Document]:
    """
    This is a set of Documents to pre-populate a graph vector store,
    with entries placed in a certain way.

    Space of the entries (under Euclidean similarity):

                      A0    (*)
        ....        AL   AR       <....
        :              |              :
        :              |  ^           :
        v              |  .           v
                       |   :
       TR              |   :          BL
    T0   --------------x--------------   B0
       TL              |   :          BR
                       |   :
                       |  .
                       | .
                       |
                    FL   FR
                      F0

    the query point is meant to be at (*).
    the A are bidirectionally with B
    the A are outgoing to T
    the A are incoming from F
    The links are like: L with L, 0 with 0 and R with R.
    """

    docs_a = [
        Document(id="AL", page_content="[-1, 9]", metadata={"label": "AL"}),
        Document(id="A0", page_content="[0, 10]", metadata={"label": "A0"}),
        Document(id="AR", page_content="[1, 9]", metadata={"label": "AR"}),
    ]
    docs_b = [
        Document(id="BL", page_content="[9, 1]", metadata={"label": "BL"}),
        Document(id="B0", page_content="[10, 0]", metadata={"label": "B0"}),
        Document(id="BR", page_content="[9, -1]", metadata={"label": "BR"}),
    ]
    docs_f = [
        Document(id="FL", page_content="[1, -9]", metadata={"label": "FL"}),
        Document(id="F0", page_content="[0, -10]", metadata={"label": "F0"}),
        Document(id="FR", page_content="[-1, -9]", metadata={"label": "FR"}),
    ]
    docs_t = [
        Document(id="TL", page_content="[-9, -1]", metadata={"label": "TL"}),
        Document(id="T0", page_content="[-10, 0]", metadata={"label": "T0"}),
        Document(id="TR", page_content="[-9, 1]", metadata={"label": "TR"}),
    ]
    for doc_a, suffix in zip(docs_a, ["l", "0", "r"]):
        doc_a.metadata["tag"] = f"ab_{suffix}"
        doc_a.metadata["out"] = f"at_{suffix}"
        doc_a.metadata["in"] = f"af_{suffix}"
    for doc_b, suffix in zip(docs_b, ["l", "0", "r"]):
        doc_b.metadata["tag"] = f"ab_{suffix}"
    for doc_t, suffix in zip(docs_t, ["l", "0", "r"]):
        doc_t.metadata["in"] = f"at_{suffix}"
    for doc_f, suffix in zip(docs_f, ["l", "0", "r"]):
        doc_f.metadata["out"] = f"af_{suffix}"
    return docs_a + docs_b + docs_f + docs_t


@pytest.fixture(scope="session")
def parser_store(
    request: pytest.FixtureRequest,
    store_factory: StoreFactory,
    graph_vector_store_docs: list[Document],
) -> StoreAdapter:
    return store_factory.create(
        request, ParserEmbeddings(dimension=2), graph_vector_store_docs
    )



from langchain_core.documents import Document
from langchain_graph_retriever.retrievers.graph_traversal_retriever import (
    GraphTraversalRetriever,
)
from langchain_graph_retriever.retrievers.strategy.eager import (
    Eager,
)
from tests.integration_tests.assertions import assert_document_format, sorted_doc_ids
from tests.integration_tests.retrievers.animal_docs import (
    ANIMALS_DEPTH_0_EXPECTED,
    ANIMALS_QUERY,
)
from tests.integration_tests.stores import StoreAdapter


async def test_animals_bidir_collection_eager(animal_store: StoreAdapter, invoker):
    # test graph-search on a normalized bi-directional edge
    retriever = GraphTraversalRetriever(
        store=animal_store,
        edges=["keywords"],
        strategy=Eager(k=100, start_k=2, max_depth=0),
    )

    docs: list[Document] = await invoker(
        retriever, ANIMALS_QUERY, strategy={"max_depth": 0}
    )
    assert sorted_doc_ids(docs) == ANIMALS_DEPTH_0_EXPECTED

    docs = await invoker(retriever, ANIMALS_QUERY, strategy={"max_depth": 1})
    assert sorted_doc_ids(docs) == [
        "cat",
        "coyote",
        "fox",
        "gazelle",
        "hyena",
        "jackal",
        "mongoose",
    ]

    docs = await invoker(retriever, ANIMALS_QUERY, strategy={"max_depth": 2})
    assert sorted_doc_ids(docs) == [
        "alpaca",
        "bison",
        "cat",
        "chicken",
        "cockroach",
        "coyote",
        "crow",
        "dingo",
        "dog",
        "fox",
        "gazelle",
        "horse",
        "hyena",
        "jackal",
        "llama",
        "mongoose",
        "ostrich",
    ]


async def test_animals_bidir_item(animal_store: StoreAdapter, invoker):
    retriever = GraphTraversalRetriever(
        store=animal_store,
        edges=["habitat"],
    )

    docs: list[Document] = await invoker(
        retriever, ANIMALS_QUERY, strategy=Eager(k=10, start_k=2, max_depth=0)
    )
    assert sorted_doc_ids(docs) == ANIMALS_DEPTH_0_EXPECTED

    docs = await invoker(
        retriever, ANIMALS_QUERY, strategy=Eager(k=10, start_k=2, max_depth=1)
    )
    assert sorted_doc_ids(docs) == [
        "bobcat",
        "cobra",
        "deer",
        "elk",
        "fox",
        "mongoose",
    ]

    docs = await invoker(
        retriever, ANIMALS_QUERY, strategy=Eager(k=10, start_k=2, max_depth=2)
    )
    assert sorted_doc_ids(docs) == [
        "bobcat",
        "cobra",
        "deer",
        "elk",
        "fox",
        "mongoose",
    ]


async def test_animals_item_to_collection(animal_store: StoreAdapter, invoker):
    retriever = GraphTraversalRetriever(
        store=animal_store,
        edges=[("habitat", "keywords")],
    )

    docs: list[Document] = await invoker(
        retriever, ANIMALS_QUERY, strategy=Eager(k=10, start_k=2, max_depth=0)
    )
    assert sorted_doc_ids(docs) == ANIMALS_DEPTH_0_EXPECTED

    docs = await invoker(
        retriever, ANIMALS_QUERY, strategy=Eager(k=10, start_k=2, max_depth=1)
    )
    assert sorted_doc_ids(docs) == ["bear", "bobcat", "fox", "mongoose"]

    docs = await invoker(
        retriever, ANIMALS_QUERY, strategy=Eager(k=10, start_k=2, max_depth=2)
    )
    assert sorted_doc_ids(docs) == ["bear", "bobcat", "caribou", "fox", "mongoose"]


async def test_parser(parser_store: StoreAdapter, invoker):
    retriever = GraphTraversalRetriever(
        store=parser_store,
        edges=[("out", "in"), "tag"],
        strategy=Eager(k=10, start_k=2, max_depth=2),
    )

    docs: list[Document] = await invoker(
        retriever, "[2, 10]", strategy=Eager(k=10, start_k=2, max_depth=0)
    )
    ss_labels = {doc.metadata["label"] for doc in docs}
    assert ss_labels == {"AR", "A0"}
    assert_document_format(docs[0])

    docs = await invoker(retriever, "[2, 10]")
    # this is a set, as some of the internals of trav.search are set-driven
    # so ordering is not deterministic:
    ts_labels = {doc.metadata["label"] for doc in docs}
    assert ts_labels == {"AR", "A0", "BR", "B0", "TR", "T0"}
    assert_document_format(docs[0])


async def test_earth(earth_store: StoreAdapter, invoker):
    retriever = GraphTraversalRetriever(
        store=earth_store,
        edges=[("outgoing", "incoming"), "keywords"],
        strategy=Eager(k=10, start_k=2, max_depth=0),
    )

    docs: list[Document] = await invoker(
        retriever, "Earth", strategy=Eager(k=10, start_k=1, max_depth=0)
    )
    assert sorted_doc_ids(docs) == ["doc2"]

    docs = await invoker(
        retriever, "Earth", strategy=Eager(k=10, start_k=2, max_depth=0)
    )
    assert sorted_doc_ids(docs) == ["doc1", "doc2"]

    docs = await invoker(
        retriever, "Earth", strategy=Eager(k=10, start_k=1, max_depth=1)
    )
    assert sorted_doc_ids(docs) == ["doc1", "doc2", "greetings"]



from typing import (
    Any,
    List,
    Sequence,
)

try:
    from langchain_community.vectorstores import Cassandra
except (ImportError, ModuleNotFoundError):
    raise ImportError("please `pip install langchain-community cassio`")

from langchain_core.documents import Document

from .base import METADATA_EMBEDDING_KEY, StoreAdapter


class CassandraStoreAdapter(StoreAdapter[Cassandra]):
    def __init__(self, vector_store: Cassandra):
        super().__init__(vector_store)

    @property
    def supports_normalized_metadata(self) -> bool:
        return False

    def similarity_search_with_embedding_by_vector(  # type: ignore
        self,
        embedding: List[float],
        k: int = 4,
        filter: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> List[Document]:
        results = self._similarity_search_with_embedding_id_by_vector(
            embedding=embedding,
            k=k,
            filter=filter,
            **kwargs,
        )
        docs: List[Document] = []
        for doc, embedding, id in results:
            doc.metadata[METADATA_EMBEDDING_KEY] = embedding
            doc.id = id
            docs.append(doc)
        return docs

    def _similarity_search_with_embedding_id_by_vector(
        self,
        embedding: list[float],
        k: int = 4,
        filter: dict[str, str] | None = None,
        body_search: str | list[str] | None = None,
    ) -> list[tuple[Document, list[float], str]]:
        """Return docs most similar to embedding vector.

        Args:
            embedding: Embedding to look up documents similar to.
            k: Number of Documents to return. Defaults to 4.
            filter: Filter on the metadata to apply.
            body_search: Document textual search terms to apply.
                Only supported by Astra DB at the moment.
        Returns:
            List of (Document, embedding, id), the most similar to the query vector.
        """
        kwargs: dict[str, Any] = {}
        if filter is not None:
            kwargs["metadata"] = filter
        if body_search is not None:
            kwargs["body_search"] = body_search

        hits = self.vector_store.table.ann_search(
            vector=embedding,
            n=k,
            **kwargs,
        )
        return [
            (
                self.vector_store._row_to_document(row=hit),
                hit["vector"],
                hit["row_id"],
            )
            for hit in hits
        ]

    async def asimilarity_search_with_embedding(
        self,
        query: str,
        k: int = 4,
        filter: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> tuple[list[float], list[Document]]:
        query_embedding = self._safe_embedding.embed_query(text=query)
        docs = await self.asimilarity_search_with_embedding_by_vector(
            embedding=query_embedding,
            k=k,
            filter=filter,
            **kwargs,
        )
        return query_embedding, docs

    async def asimilarity_search_with_embedding_by_vector(  # type: ignore
        self, **kwargs: Any
    ) -> list[Document]:
        results = (
            await self.vector_store.asimilarity_search_with_embedding_id_by_vector(
                **kwargs
            )
        )
        docs: List[Document] = []
        for doc, embedding, id in results:
            doc.metadata[METADATA_EMBEDDING_KEY] = embedding
            doc.id = id
            docs.append(doc)
        return docs

    def get(self, ids: Sequence[str], /, **kwargs: Any) -> list[Document]:
        """Get documents by id."""
        docs: list[Document] = []
        for id in ids:
            doc = self.vector_store.get_by_document_id(id, **kwargs)
            if doc is not None:
                docs.append(doc)
        return docs

    async def aget(self, ids: Sequence[str], /, **kwargs: Any) -> list[Document]:
        """Get documents by id."""
        docs: list[Document] = []
        for id in ids:
            doc = await self.vector_store.aget_by_document_id(id, **kwargs)
            if doc is not None:
                docs.append(doc)
        return docs



from langchain_core.documents import Document
from langchain_core.vectorstores import InMemoryVectorStore
from langchain_graph_retriever.retrievers.graph_traversal_retriever import (
    GraphTraversalRetriever,
)
from langchain_graph_retriever.retrievers.store_adapters.in_memory import (
    InMemoryStoreAdapter,
)
from langchain_graph_retriever.retrievers.strategy.mmr import (
    Mmr,
)
from tests.embeddings.simple_embeddings import Angular2DEmbeddings
from tests.integration_tests.assertions import sorted_doc_ids
from tests.integration_tests.retrievers.animal_docs import (
    ANIMALS_DEPTH_0_EXPECTED,
    ANIMALS_QUERY,
)
from tests.integration_tests.stores import StoreAdapter


async def test_animals_bidir_collection(animal_store: StoreAdapter, invoker):
    # test graph-search on a normalized bi-directional edge
    retriever = GraphTraversalRetriever(
        store=animal_store,
        edges=["keywords"],
    )

    docs: list[Document] = await invoker(
        retriever, ANIMALS_QUERY, strategy=Mmr(k=4, start_k=2, max_depth=0)
    )
    assert sorted_doc_ids(docs) == ANIMALS_DEPTH_0_EXPECTED

    docs = await invoker(
        retriever, ANIMALS_QUERY, strategy=Mmr(k=4, start_k=2, max_depth=1)
    )
    assert sorted_doc_ids(docs) == ["cat", "gazelle", "hyena", "mongoose"]

    docs = await invoker(
        retriever, ANIMALS_QUERY, strategy=Mmr(k=6, start_k=2, max_depth=2)
    )
    assert sorted_doc_ids(docs) == [
        "bison",
        "cat",
        "fox",
        "gazelle",
        "hyena",
        "mongoose",
    ]


async def test_animals_bidir_item(animal_store: StoreAdapter, invoker):
    retriever = GraphTraversalRetriever(
        store=animal_store,
        edges=["habitat"],
    )

    docs: list[Document] = await invoker(
        retriever, ANIMALS_QUERY, strategy=Mmr(k=10, start_k=2, max_depth=0)
    )
    assert sorted_doc_ids(docs) == ANIMALS_DEPTH_0_EXPECTED

    docs = await invoker(
        retriever, ANIMALS_QUERY, strategy=Mmr(k=10, start_k=2, max_depth=1)
    )
    assert sorted_doc_ids(docs) == [
        "bobcat",
        "cobra",
        "deer",
        "elk",
        "fox",
        "mongoose",
    ]

    docs = await invoker(
        retriever, ANIMALS_QUERY, strategy=Mmr(k=10, start_k=2, max_depth=2)
    )
    assert sorted_doc_ids(docs) == [
        "bobcat",
        "cobra",
        "deer",
        "elk",
        "fox",
        "mongoose",
    ]


async def test_animals_item_to_collection(animal_store: StoreAdapter, invoker):
    retriever = GraphTraversalRetriever(
        store=animal_store,
        edges=[("habitat", "keywords")],
    )

    docs: list[Document] = await invoker(
        retriever, ANIMALS_QUERY, strategy=Mmr(k=10, start_k=2, max_depth=0)
    )
    assert sorted_doc_ids(docs) == ANIMALS_DEPTH_0_EXPECTED

    docs = await invoker(
        retriever, ANIMALS_QUERY, strategy=Mmr(k=10, start_k=2, max_depth=1)
    )
    assert sorted_doc_ids(docs) == ["bear", "bobcat", "fox", "mongoose"]

    docs = await invoker(
        retriever, ANIMALS_QUERY, strategy=Mmr(k=10, start_k=2, max_depth=2)
    )
    assert sorted_doc_ids(docs) == ["bear", "bobcat", "caribou", "fox", "mongoose"]


async def test_traversal_mem(invoker) -> None:
    """ Test end to end construction and MMR search.
    The embedding function used here ensures `texts` become
    the following vectors on a circle (numbered v0 through v3):

           ______ v2
          //      \\
         //        \\  v1
    v3  ||    .     || query
         \\        //  v0
          \\______//                 (N.B. very crude drawing)

    With start_k==2 and k==2, when query is at (1, ),
    one expects that v2 and v0 are returned (in some order)
    because v1 is "too close" to v0 (and v0 is closer than v1)).

    Both v2 and v3 are reachable via edges from v0, so once it is
    selected, those are both considered.
    """
    v0 = Document(id="v0", page_content="-0.124")
    v1 = Document(id="v1", page_content="+0.127")
    v2 = Document(id="v2", page_content="+0.25")
    v3 = Document(id="v3", page_content="+1.0")

    v0.metadata["outgoing"] = "link"
    v2.metadata["incoming"] = "link"
    v3.metadata["incoming"] = "link"

    store = InMemoryVectorStore(embedding=Angular2DEmbeddings())
    store.add_documents([v0, v1, v2, v3])

    strategy = Mmr(k=2, start_k=2, max_depth=2)
    retriever = GraphTraversalRetriever(
        store=InMemoryStoreAdapter(
            vector_store=store, support_normalized_metadata=False
        ),
        edges=[("outgoing", "incoming")],
        strategy=strategy,
    )

    docs: list[Document] = await invoker(retriever, "0.0")
    assert sorted_doc_ids(docs) == ["v0", "v2"]

    # With max depth 0, no edges are traversed, so this doesn't reach v2 or v3.
    # So it ends up picking "v1" even though it's similar to "v0".
    docs = await invoker(retriever, "0.0", strategy={"max_depth": 0})
    assert sorted_doc_ids(docs) == ["v0", "v1"]

    # With max depth 0 but higher `start_k`, we encounter v2
    docs = await invoker(retriever, "0.0", strategy={"start_k": 3, "max_depth": 0})
    assert sorted_doc_ids(docs) == ["v0", "v2"]

    # v0 score is .46, v2 score is 0.16 so it won't be chosen.
    docs = await invoker(retriever, "0.0", strategy={"score_threshold": 0.2})
    assert sorted_doc_ids(docs) == ["v0"]

    # with k=4 we should get all of the documents.
    docs = await invoker(retriever, "0.0", strategy={"k": 4})
    assert sorted_doc_ids(docs) == ["v0", "v1", "v2", "v3"]



from typing import Any, Sequence

from langchain_core.documents import BaseDocumentTransformer, Document


class KeybertKeywordExtractor(BaseDocumentTransformer):
    """Add metadata to documents about keywords using `KeyBERT <https://maartengr.github.io/KeyBERT/>`_.

    KeyBERT is a minimal and easy-to-use keyword extraction technique that
    leverages BERT embeddings to create keywords and keyphrases that are most
    similar to a document.

    The KeybertKeywordExtractor uses KeyBERT add a list of keywords to a
    document's metadata.

    Preliminaries
    -------------

    Install the ``keybert`` package.

    Note that ``bs4`` is also installed to support the WebBaseLoader in the example,
    but not needed by the KeybertKeywordExtractor itself.

    .. code-block:: bash

        pip install -q langchain_community bs4 keybert

    Example
    -----

    We load the ``state_of_the_union.txt`` file, chunk it, then for each chunk we
    add keywords to the metadata.

    .. code-block:: python

        from langchain_community.document_loaders import WebBaseLoader
        from langchain_community.document_transformers import KeybertKeywordExtractor
        from langchain_text_splitters import CharacterTextSplitter

        loader = WebBaseLoader(
            "https://raw.githubusercontent.com/hwchase17/chat-your-data/master/state_of_the_union.txt"
        )
        raw_documents = loader.load()

        text_splitter = CharacterTextSplitter(chunk_size=1000, chunk_overlap=0)
        documents = text_splitter.split_documents(raw_documents)

        extractor = KeybertKeywordExtractor()
        documents = extractor.transform_documents(documents)

        print(documents[0].metadata)

    .. code-block:: output

        {'source': 'https://raw.githubusercontent.com/hwchase17/chat-your-data/master/state_of_the_union.txt', 'keywords': ['putin', 'vladimir', 'ukrainian', 'russia', 'ukraine']}

    Args:
        batch_size: The number of documents to process in each batch (default ``8``)
        metadata_key: The name of the key used in the metadata output (default ``keywords``)
        model: The KeyBERT model to use. (default ``all-MiniLM-L6-v2``)
    """  # noqa: E501

    def __init__(
        self,
        *,
        batch_size: int = 8,
        metadata_key: str = "keywords",
        model: str = "all-MiniLM-L6-v2",
    ):
        try:
            import keybert  # type: ignore

            self._kw_model = keybert.KeyBERT(model=model)
        except ImportError:
            raise ImportError(
                "keybert is required for the KeybertLinkExtractor. "
                "Please install it with `pip install keybert`."
            ) from None

        self._batch_size = batch_size
        self._metadata_key = metadata_key

    def transform_documents(
        self, documents: Sequence[Document], **kwargs: Any
    ) -> Sequence[Document]:
        """Adds keyword metadata to documents using KeyBERT.

        Args:
            documents: The sequence of documents to transform
            kwargs: Keyword arguments to pass to KeyBERT. See: https://maartengr.github.io/KeyBERT/api/keybert.html#keybert._model.KeyBERT.extract_keywords
        """
        for i in range(0, len(documents), self._batch_size):
            batch = documents[i : i + self._batch_size]
            texts = [item.page_content for item in batch]
            extracted = self._kw_model.extract_keywords(docs=texts, **kwargs)
            if len(texts) == 1:
                # Even though we pass a list, if it contains one item, keybert will
                # flatten it. This means it's easier to just call the special case
                # for one item.
                batch[0].metadata[self._metadata_key] = [kw[0] for kw in extracted]
            else:
                for i, keywords in enumerate(extracted):
                    batch[i].metadata[self._metadata_key] = [kw[0] for kw in keywords]
        return documents



from typing import (
    Any,
    Dict,
    List,
    Optional,
    Sequence,
)

try:
    from langchain_community.vectorstores import OpenSearchVectorSearch
except (ImportError, ModuleNotFoundError):
    raise ImportError("please `pip install langchain-community opensearch-py`")

from langchain_core.documents import Document

from .base import METADATA_EMBEDDING_KEY, StoreAdapter


class OpenSearchStoreAdapter(StoreAdapter[OpenSearchVectorSearch]):
    def __init__(self, vector_store: OpenSearchVectorSearch):
        super().__init__(vector_store)
        if self.vector_store.engine not in ["lucene", "faiss"]:
            msg = (
                f"Invalid engine for Traversal: '{self.vector_store.engine}'"
                " please instantiate the Open Search Vector Store with"
                " either the 'lucene' or 'faiss' engine"
            )
            raise ValueError(msg)

    @property
    def supports_normalized_metadata(self) -> bool:
        return True

    def _build_filter(
        self, filter: Optional[Dict[str, str]] = None
    ) -> List[Dict[str, Any]] | None:
        if filter is None:
            return None
        return [
            {
                "terms" if isinstance(value, list) else "term": {
                    f"metadata.{key}.keyword": value
                }
            }
            for key, value in filter.items()
        ]

    def similarity_search_with_embedding_by_vector(
        self,
        embedding: List[float],
        k: int = 4,
        filter: Optional[Dict[str, str]] = None,
        **kwargs: Any,
    ) -> List[Document]:
        """Returns docs (with embeddings) most similar to the query vector."""
        if filter is not None:
            # use an efficient_filter to collect results that
            # are near the embedding vector until up to 'k'
            # documents that match the filter are found.
            kwargs["efficient_filter"] = {
                "bool": {"must": self._build_filter(filter=filter)}
            }

        docs = self.vector_store.similarity_search_by_vector(
            embedding=embedding,
            k=k,
            metadata_field="*",
            **kwargs,
        )

        # if metadata=="*" on the search, then the document
        # embedding vector and text are included in the
        # document metadata in the returned document.
        #
        # The actual document metadata is moved down into a
        # sub "metadata" key.
        for doc in docs:
            embedding = doc.metadata["vector_field"]
            doc.metadata = doc.metadata["metadata"] or {}
            doc.metadata[METADATA_EMBEDDING_KEY] = embedding

        return docs

    def get(self, ids: Sequence[str], /, **kwargs: Any) -> list[Document]:
        """Get documents by id."""
        try:
            from opensearchpy.exceptions import NotFoundError
        except (ImportError, ModuleNotFoundError):
            msg = "please `pip install opensearch-py`."
            raise ImportError(msg)

        docs: list[Document] = []
        for id in ids:
            try:
                hit = self.vector_store.client.get(
                    index=self.vector_store.index_name,
                    id=id,
                    _source_includes=["text", "metadata"],
                    **kwargs,
                )
                docs.append(
                    Document(
                        page_content=hit["_source"]["text"],
                        metadata=hit["_source"]["metadata"],
                        id=hit["_id"],
                    )
                )
            except NotFoundError:
                pass
        return docs
